# check_index_found(index_id, identifier, dimension_id) {
#   if (length(index_id) == 0)
#     stop(paste0("identifier '", identifier, "' not found in the dimension '",
#       dimension_id, "'"))
# }

# STATION_DIMNAME = "station"
# LEAD_TIME_DIMNAME = "lead_time"
# TIME_DIMNAME = "time"
# ENS_MEMBER_DIMNAME = "ens_member"
# STR_LEN_DIMNAME = "str_len"

# # int station_id[station]
# STATION_ID_VARNAME = "station_id"
# # char station_name[str_len,station]
# STATION_NAME_VARNAME = "station_name"
# # float lat[station]
# LAT_VARNAME = LAT_VARNAME
# # float lon[station]
# lon_varname = "lon"
# # float x[station]
# x_varname = "x"
# # float y[station]
# y_varname = "y"
# # float area[station]
# area_varname = "area"
# # float elevation[station]
# elevation_varname = "elevation"

# conventional_varnames = c(
#   STATION_DIMNAME ,
#   LEAD_TIME_DIMNAME ,
#   TIME_DIMNAME ,
#   ENS_MEMBER_DIMNAME ,
#   STR_LEN_DIMNAME ,
#   STATION_ID_VARNAME ,
#   STATION_NAME_VARNAME ,
#   LAT_VARNAME ,
#   lon_varname ,
#   x_varname ,
#   y_varname ,
#   area_varname ,
#   elevation_varname
# )

# mandatory_global_attributes = c("title", INSTITUTION_ATTR_KEY, SOURCE_ATTR_KEY, CATCHMENT_ATTR_KEY, COMMENT_ATTR_KEY)


# get_default_dim_order() {
#   return(c(LEAD_TIME_DIMNAME, STATION_DIMNAME, ENS_MEMBER_DIMNAME, TIME_DIMNAME))
# }

# splice_named_var(d, ncdims = character()) {
#   default_order = get_default_dim_order()
#   d = as.integer(d)
#   stopifnot(length(d) == 4)
#   stopifnot(is.vector(d))
#   # lead_time,station,ens_member,time
#   names(d) = default_order
#   if (length(ncdims) > 0) {
#     if (!all(ncdims %in% default_order)) {
#       stop(paste0("Invalid dimensions for a data variable: ", paste(ncdims,
#         collapse = ",")))
#     } else {
#       d = d[ncdims]
#       names(d) = ncdims
#     }
#   }
#   return(d)
# }


# dim_names(x) {
#   attr(x, 'dim_names')
# }

# "dim_names<-"(x, value) {
#   d = dim(x)
#   if(is.array(x)) {
#     if(length(d) != length(value)) stop("dim names is not equal to the number of dimensions of the array")
#     if(length(unique(value)) != length(d)) stop("specified dim names are not unique")
#   } else if (is.vector(x)){
#     stopifnot(length(value) == 1)
#   } else { stop('not an array nor a vector - cannot set dim_names')}
#   attr(x, 'dim_names') = value
#   return(x)
# }

# reduce_dimensions(x, subset_dim_names){
#   dimsize_input = dim(x)
#   dn = dim_names(x)
#   if(is.null(dn)) stop('the input array must have a valid dim_names attribute')
#   if(length(dn) != length(dimsize_input)) stop('the input array and its dim_names attribute are differing in length')

#   names(dimsize_input) = dn
#   if(missing(subset_dim_names) || is.na(subset_dim_names))
#     subset_dim_names = dn[dimsize_input > 1]

#   diffdim = setdiff(subset_dim_names, dn)
#   if (length(diffdim)>0) stop(paste0('Dimension names to slice but not found in array dim names: ', paste(diffdim, collpase=', ')))

#   dropped_dims = setdiff(dn,subset_dim_names)
#   if( any(dimsize_input[dropped_dims] > 1)) stop('Cannot drop non-degenerate when subsetting')

#   w = match(subset_dim_names,dn)
#   other = match(setdiff(dn, subset_dim_names),dn)

#   x_reordered = aperm(x, c(w, other))

#   reordered_dim_names = dn[c(w, other)]
#   reordered_dim_sizes = dim(x_reordered)

#   new_dim_sizes = reordered_dim_sizes[1:length(w)]
#   new_dim_names = reordered_dim_names[1:length(w)]

#   y = drop(x_reordered)
#   # We want however to maintain degenerate
#   # dimensions that have been explicitly asked for,
#   # and that would have been otherwise dropped
#   y = array(y, new_dim_sizes)

#   dim_names(y) = new_dim_names
#   return(y)
# }

#' @import magrittr
from typing import Any, Dict, Tuple

import xarray as xr

from efts_io.conventions import FILLVALUE_ATTR_KEY, UNITS_ATTR_KEY


def create_data_variable(data_var_def: Dict[str, Any], dimensions: Tuple[str, Tuple]) -> xr.Variable:
    import numpy as np
    import xarray as xr

    a = data_var_def
    #    (c("name", UNITS_ATTR_KEY) %in% names(a)) %>% all %>% stopifnot
    varname = a["name"]
    longname = a.get("longname", varname)
    precision = a.get("precision", "double")
    missval = a.get("missval", -9999)

    dimnames = [d[0] for d in dimensions]
    if not isinstance(dimnames[0], str):
        raise TypeError("Dimension names must be strings.")
    shape = tuple(len(d[1]) for d in dimensions)
    return xr.Variable(
        dims=dimnames,
        data=np.empty(shape, dtype=float),  # TODO: should this use precision?
        encoding={FILLVALUE_ATTR_KEY: missval},
        attrs={
            "longname": longname,
            UNITS_ATTR_KEY: a[UNITS_ATTR_KEY],
            "missval": missval,
            "precision": precision,
        },
    )

    # xr.Variable(dims=dimensions, data, attrs=None, encoding=None, fastpath=False)
    # vardef = ncdf4::ncvar_def(name = varname, units = a[UNITS_ATTR_KEY], dim = dimensions,
    # longname = ifelse("longname" %in% names(a), a["longname"], varname)
    # precision = ifelse("precision" %in% names(a), a["precision"], "double")
    # missval = ifelse("missval" %in% names(a), a["missval"]], -9999)
    # vardef = ncdf4::ncvar_def(name = varname, units = a[UNITS_ATTR_KEY], dim = dimensions,
    #     missval = missval, longname = longname, prec = precision)



# import netCDF4
import numpy as np
import pandas as pd
import pytest
from efts_io._ncdf_stf2 import StfDataType, StfVariable
from efts_io.wrapper import EftsDataSet, template_variable_attributes, xr_efts


def test_create_new_efts_stf2():
    """Placeholder test, for upcoming version of the conventions.

    Note the use of integers for station IDs, not strings.
    """
    import efts_io.wrapper as wrap

    issue_times = pd.date_range("2010-01-01", periods=31, freq="D")
    station_ids = [123, 456]
    lead_times = np.arange(start=1, stop=4, step=1)
    lead_time_tstep = "hours"
    ensemble_size = 10
    station_names = [f"{x} station name" for x in station_ids]
    nc_attributes = None
    latitudes = None
    longitudes = None
    areas = None

    def _create_test_ds():
        d = wrap.xr_efts(
            issue_times,
            station_ids,
            lead_times,
            lead_time_tstep,
            ensemble_size,
            station_names,
            nc_attributes,
            latitudes,
            longitudes,
            areas,
        )
        return EftsDataSet(d)

    # NOTE: should it be? is it wise to allow missing values for mandatory variables
    w = _create_test_ds()

    w.create_data_variables(
        {
            "rain_obs": {
                "name": "rain_obs",
                "longname": "Observed Rainfall Amount",
                "units": "mm",
                "dim_type": "4",
                "comment": "This is a comment for the variable",
                "missval": np.nan,
                "precision": "double",
                "attributes": {},
            },
        }
    )
    assert w.writeable_to_stf2()
    # create a temporary file
    import tempfile

    with tempfile.NamedTemporaryFile(suffix=".nc") as tmp:
        filename = tmp.name
        w.save_to_stf2(
            path=filename,
            variable_name="rain_obs",
            var_type=StfVariable.RAINFALL,
            data_type=StfDataType.DERIVED_FROM_OBSERVATIONS,
            ens=False,
            timestep="hours",
            data_qual=None,
        )
        # round trip test
        ds = EftsDataSet(filename)
        assert "rain_obs" in ds.data.data_vars


def test_create_new_efts_future_station_ids():
    """Placeholder test, for upcoming version of the conventions.

    Note the use of strings for station IDs, not integers.
    """
    import efts_io.wrapper as wrap

    issue_times = pd.date_range("2010-01-01", periods=31, freq="D")
    string_station_ids = ["a", "b"]
    lead_times = np.arange(start=1, stop=4, step=1)
    lead_time_tstep = "hours"
    ensemble_size = 10
    station_names = [f"{x} station name" for x in string_station_ids]
    nc_attributes = None
    latitudes = None
    longitudes = None
    areas = None

    def _create_test_ds():
        d = wrap.xr_efts(
            issue_times,
            string_station_ids,
            lead_times,
            lead_time_tstep,
            ensemble_size,
            station_names,
            nc_attributes,
            latitudes,
            longitudes,
            areas,
        )
        return EftsDataSet(d)

    # NOTE: should it be? is it wise to allow missing values for mandatory variables
    w = _create_test_ds()

    w.create_data_variables(
        {
            "rain_obs": {
                "name": "rain_obs",
                "longname": "Observed Rainfall Amount",
                "units": "mm",
                "dim_type": "4",
                "comment": "This is a comment for the variable",
                "missval": np.nan,
                "precision": "double",
                "attributes": {},
            },
        }
    )
    # 2025-11 It has been decided to support transparent conversion of string station
    # IDs to integers on save for STF2.0.
    # It was otherwise confusing for users.
    # besides it helps to promote the use of string station IDs in memory datasets.
    assert w.writeable_to_stf2()
    # if w.writeable_to_stf2():
    #     raise RuntimeError("This should not be flagged as writeable to STF2.0, station IDs are strings.")


def test_repro_issue_16():
    """Try to repro as closely as possible the issue reported in #16."""
    station_ids = [1, 2, 3]
    _saving_to_stf2(station_ids)


def test_large_station_integers():
    """Try to repro as closely as possible the issue reported in #17."""
    station_ids = [1, 2, 123456789123]
    with pytest.raises(OverflowError):
        _saving_to_stf2(station_ids, intdata_type="i4", delete=False)
    _saving_to_stf2(station_ids, intdata_type="i8", delete=True)


def _saving_to_stf2(station_ids, intdata_type="i4", delete=True):
    xr_ds = xr_efts(
        issue_times=pd.date_range("2023-10-01", periods=31, freq="D"),
        station_ids=station_ids,
        lead_times=np.arange(start=1, stop=4, step=1),
        lead_time_tstep="hours",
        ensemble_size=1,
        station_names=["station_1", "station_2", "station_3"],
        nc_attributes={
            "title": "Test dataset",
            "institution": "Test institution",
            "source": "Test source",
            "history": "Created for testing purposes",
            "references": "None",
            "comment": "This is a test dataset",
            "catchment": "Test catchment",
            # "STF_convention_version" can be skipped in memoy, this is added on save to stf2
            # "STF_nc_spec" can be skipped in memoy, this is added on save to stf2
        },
        areas=[10.0, 20.0, 30.0],
        latitudes=[60.0, 61.0, 62.0],
        longitudes=[10.0, 11.0, 12.0],
    )
    eds = EftsDataSet(xr_ds)
    eds.stf2_int_datatype = intdata_type
    eds.create_data_variables(
        {
            "rain_obs": {
                "name": "rain_obs",
                "longname": "Observed Rainfall Amount",
                "units": "mm",
                "dim_type": "4",
                "comment": "This is a comment for the variable",
                "missval": np.nan,
                "precision": "double",
                "attributes": {},
            },
        }
    )
    assert eds.writeable_to_stf2()
    # create a temporary file
    import tempfile

    # save to STF2.0 will clean up the file if write fails,
    # so in that case we should allow for deletion to be True or False
    with tempfile.NamedTemporaryFile(suffix=".nc", delete=delete) as tmp:
        filename = tmp.name
        eds.save_to_stf2(
            path=filename,
            variable_name="rain_obs",
            var_type=StfVariable.RAINFALL,
            data_type=StfDataType.DERIVED_FROM_OBSERVATIONS,
            ens=False,
            timestep="hours",
            data_qual=None,
        )
        # round trip test
        ds = EftsDataSet(filename)
        assert "rain_obs" in ds.data.data_vars


def test_new_variable():
    """Test the new_variable method of EftsDataSet."""
    import efts_io.wrapper as wrap

    # Create a base dataset using existing test helper pattern
    issue_times = pd.date_range("2010-01-01", periods=10, freq="D")
    station_ids = [123, 456]
    lead_times = np.arange(start=1, stop=4, step=1)
    lead_time_tstep = "hours"
    ensemble_size = 5

    d = wrap.xr_efts(
        issue_times=issue_times,
        station_ids=station_ids,
        lead_times=lead_times,
        lead_time_tstep=lead_time_tstep,
        ensemble_size=ensemble_size,
    )
    eds = EftsDataSet(d)

    # Test 1: Create a variable with automatic NaN initialization
    var_attrs = {
        "units": "mm",
        "long_name": "Test Variable",
        "comment": "This is a test variable",
    }
    new_var = eds.new_variable(
        varname="test_var",
        dim_names=["station_id", "time"],
        var_attributes=var_attrs,
    )

    assert "test_var" in eds.data.variables
    assert new_var.name == "test_var"
    assert new_var.dims == ("station_id", "time")
    assert new_var.shape == (len(station_ids), len(issue_times))
    assert np.all(np.isnan(new_var.values))
    assert new_var.attrs["units"] == "mm"
    assert new_var.attrs["long_name"] == "Test Variable"

    # Test 2: Create a variable with explicit data
    data_shape = (len(station_ids), len(issue_times))
    test_data = np.random.randn(*data_shape)
    var_attrs_2 = {
        "units": "degC",
        "long_name": "Temperature",
    }
    new_var_2 = eds.new_variable(
        varname="temperature",
        dim_names=["station_id", "time"],
        var_attributes=var_attrs_2,
        data=test_data,
    )

    assert "temperature" in eds.data.variables
    assert new_var_2.shape == data_shape
    assert np.allclose(new_var_2.values, test_data)
    assert new_var_2.attrs["units"] == "degC"

    # Test 3: Create a 4D variable (forecast-like dimensions)
    var_attrs_3 = {
        "units": "mm/h",
        "long_name": "Rainfall Rate Forecast",
    }
    new_var_3 = eds.new_variable(
        varname="rain_fcast",
        dim_names=["lead_time", "station_id", "realisation", "time"],
        var_attributes=var_attrs_3,
    )

    assert "rain_fcast" in eds.data.variables
    assert new_var_3.dims == ("lead_time", "station_id", "realisation", "time")
    expected_shape = (
        len(lead_times),
        len(station_ids),
        ensemble_size,
        len(issue_times),
    )
    assert new_var_3.shape == expected_shape

    # Test 4: Error when variable already exists
    with pytest.raises(ValueError, match="already exists"):
        eds.new_variable(
            varname="test_var",
            dim_names=["station_id", "time"],
            var_attributes=var_attrs,
        )

    # Test 5: Error when units attribute is missing
    with pytest.raises(ValueError, match="must include 'units'"):
        eds.new_variable(
            varname="no_units_var",
            dim_names=["station_id", "time"],
            var_attributes={"long_name": "No units"},
        )

    # Test 6: Error when data shape doesn't match dimensions
    wrong_shape_data = np.random.randn(5, 5)  # Wrong shape
    with pytest.raises(ValueError, match="does not match expected shape"):
        eds.new_variable(
            varname="wrong_shape_var",
            dim_names=["station_id", "time"],
            var_attributes={"units": "m"},
            data=wrong_shape_data,
        )

    # Test 7: Use template attributes as starting point
    template_attrs = template_variable_attributes()
    template_attrs["units"] = "m3/s"
    template_attrs["long_name"] = "Streamflow"
    new_var_4 = eds.new_variable(
        varname="streamflow",
        dim_names=["station_id", "time"],
        var_attributes=template_attrs,
    )

    assert "streamflow" in eds.data.variables
    assert new_var_4.attrs["units"] == "m3/s"
    assert new_var_4.attrs["long_name"] == "Streamflow"


if __name__ == "__main__":
    # test_read_thing()
    test_create_new_efts()



from datetime import datetime
import pytest
import pandas as pd
import xarray as xr
import numpy as np
from efts_io._ncdf_stf2 import _create_cf_time_axis
from efts_io.conventions import convert_to_datetime64_utc


def test_create_cf_time_axis_valid_input():
    # Create a sample DataArray with a time dimension
    dates = pd.date_range(start="2023-01-01", periods=5, freq="D")
    data = xr.DataArray(np.random.rand(5), dims=["time"], coords={"time": dates})

    # Test with a valid time step
    result, units, calendar = _create_cf_time_axis(data, "days")

    # Check if the result is a numpy array
    assert isinstance(result, np.ndarray)
    assert len(result) == 5
    assert units == "days since 2023-01-01 00:00:00+00:00"
    assert calendar == "proleptic_gregorian"


def test_create_cf_time_axis_empty_data():
    # Create an empty DataArray
    data = xr.DataArray([], dims=["time"])

    # Test with an empty DataArray
    with pytest.raises(ValueError, match="Cannot create CF time axis from empty data array."):
        _create_cf_time_axis(data, "days")


def test_create_cf_time_axis_invalid_time_type():
    # Create a DataArray with invalid time type
    data = xr.DataArray([1, 2, 3], dims=["time"], coords={"time": [1, 2, 3]})

    # Test with invalid time type
    with pytest.raises(
        TypeError,
        match="Expected data\\[TIME_DIMNAME\\] to be of a type convertible to pd.Timestamp, got <class 'numpy.int64'> instead.",
    ):
        _create_cf_time_axis(data, "days")


# Unit tests
def test_convert_to_datetime64_utc():
    # Test with a timezone-naive pd.Timestamp
    naive_timestamp = pd.Timestamp("2023-10-01 12:00:00")
    assert convert_to_datetime64_utc(naive_timestamp) == np.datetime64("2023-10-01T12:00:00.000000000")

    # Test with a timezone-aware pd.Timestamp
    aware_timestamp = pd.Timestamp("2023-10-01 12:00:00", tz="America/New_York")
    assert convert_to_datetime64_utc(aware_timestamp) == np.datetime64("2023-10-01T16:00:00.000000000")

    # Test with a timezone-naive datetime
    naive_datetime = datetime(2023, 10, 1, 12, 0, 0)
    assert convert_to_datetime64_utc(naive_datetime) == np.datetime64("2023-10-01T12:00:00.000000000")

    # Test with a timezone-aware datetime
    from zoneinfo import ZoneInfo

    utc_tz = ZoneInfo("UTC")
    aware_datetime = datetime(2023, 10, 1, 12, 0, 0, tzinfo=utc_tz)
    assert convert_to_datetime64_utc(aware_datetime) == np.datetime64("2023-10-01T12:00:00.000000000")

    # Test with a string representation
    naive_string = "2023-10-01 12:00:00"
    assert convert_to_datetime64_utc(naive_string) == np.datetime64("2023-10-01T12:00:00.000000000")

    # Test with a timezone-aware string representation
    aware_string = "2023-10-01 12:00:00-04:00"
    assert convert_to_datetime64_utc(aware_string) == np.datetime64("2023-10-01T16:00:00.000000000")

    # Test with an np.datetime64
    naive_np_datetime = np.datetime64("2023-10-01T12:00:00.000000000")
    assert convert_to_datetime64_utc(naive_np_datetime) == np.datetime64("2023-10-01T12:00:00.000000000")


if __name__ == "__main__":
    test_create_cf_time_axis_invalid_time_type()
    test_create_cf_time_axis_empty_data()
    test_create_cf_time_axis_valid_input()



"""A thin wrapper around xarray for reading and writing Ensemble Forecast Time Series (EFTS) data sets."""

import os
from datetime import datetime
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

# import netCDF4
import numpy as np
import pandas as pd
import xarray as xr

from efts_io._ncdf_stf2 import StfDataType, StfVariable
from efts_io.conventions import (
    AREA_VARNAME,
    AXIS_ATTR_KEY,
    CATCHMENT_ATTR_KEY,
    COMMENT_ATTR_KEY,
    ENS_MEMBER_DIMNAME,
    FILLVALUE_ATTR_KEY,
    HISTORY_ATTR_KEY,
    INSTITUTION_ATTR_KEY,
    LAT_VARNAME,
    LEAD_TIME_DIMNAME,
    LON_VARNAME,
    LONG_NAME_ATTR_KEY,
    REALISATION_DIMNAME,
    SOURCE_ATTR_KEY,
    STANDARD_NAME_ATTR_KEY,
    STATION_DIMNAME,
    STATION_ID_DIMNAME,
    STATION_ID_VARNAME,
    STATION_NAME_VARNAME,
    STF_2_0_URL,
    STF_CONVENTION_VERSION_ATTR_KEY,
    STF_NC_SPEC_ATTR_KEY,
    TIME_DIMNAME,
    TIME_STANDARD_ATTR_KEY,
    TITLE_ATTR_KEY,
    UNITS_ATTR_KEY,
    ConvertibleToTimestamp,
    check_index_found,
)
from efts_io.dimensions import cftimes_to_pdtstamps
from efts_io.variables import create_efts_variables


def byte_to_string(x: Union[int, bytes]) -> str:
    """Convert a byte to a string."""
    if isinstance(x, int):
        if x > 255 or x < 0:  # noqa: PLR2004
            raise ValueError("Integer value to bytes: must be in range [0-255]")
        x = x.to_bytes(1, "little")
    if not isinstance(x, bytes):
        raise TypeError(f"Cannot cast type {type(x)} to bytes")
    return str(x, encoding="UTF-8")


def byte_array_to_string(x: np.ndarray) -> str:
    """Convert a byte array to a string."""
    s = "".join([byte_to_string(s) for s in x])
    return s.strip()


def byte_stations_to_str(byte_names: np.ndarray) -> np.ndarray:
    """Convert byte array of station names to string array."""
    return np.array([byte_array_to_string(x) for x in byte_names])


def _first_where(condition: np.ndarray) -> int:
    """Return the first index where the condition is true."""
    x = np.where(condition)[0]
    if len(x) < 1:
        raise ValueError("first_where: Invalid condition, no element is true")
    return x[0]


def load_from_stf2_file(file_path: str, time_zone_timestamps: bool) -> xr.Dataset:  # noqa: FBT001
    """Load data from an STF 2.0 netcdf file to an xarray representation.

    Args:
        file_path (str): file path
        time_zone_timestamps (bool): should we try to recognise the time zone and include it in each xarray time stamp?

    Returns:
        _type_: xarray Dataset
    """
    from xarray.coding import times

    # work around https://jira.csiro.au/browse/WIRADA-635
    # lead_time can be a problem with xarray, so do not decode "times"
    # we also use mask_and_scale=False because of https://github.com/csiro-hydroinformatics/efts-io/issues/24
    x = xr.open_dataset(file_path, decode_times=False, mask_and_scale=False)
    # replace the time and station names coordinates values
    # TODO This is probably not a long term solution for round-tripping a read/write or vice and versa
    decod = times.CFDatetimeCoder(use_cftime=True)
    var = xr.as_variable(x.coords[TIME_DIMNAME])
    time_zone = var.attrs[TIME_STANDARD_ATTR_KEY]
    time_coords = decod.decode(var, name=TIME_DIMNAME)
    tz = time_zone if time_zone_timestamps else None
    time_coords.values = cftimes_to_pdtstamps(
        time_coords.values,
        tz_str=tz,
    )
    # stat_coords = x.coords[self.STATION_DIMNAME]
    # see the use of astype later on in variable transfer, following line not needed.
    # station_names = byte_stations_to_str(x[STATION_NAME_VARNAME].values).astype(np.str_)
    # Use np.atleast_1d to handle the case where there's only one station/ensemble/lead_time,
    # which would otherwise result in a 0-dimensional array that xarray cannot use as a coordinate.
    station_ids_strings = np.atleast_1d(x[STATION_ID_VARNAME].values.astype(np.str_))
    ens_member_values = np.atleast_1d(x[ENS_MEMBER_DIMNAME].values)
    lead_time_values = np.atleast_1d(x[LEAD_TIME_DIMNAME].values)
    # x = x.assign_coords(
    #     {TIME_DIMNAME: time_coords, self.STATION_DIMNAME: station_names},
    # )

    # Create a new dataset with the desired structure
    new_dataset = xr.Dataset(
        coords={
            REALISATION_DIMNAME: (REALISATION_DIMNAME, ens_member_values),
            LEAD_TIME_DIMNAME: (LEAD_TIME_DIMNAME, lead_time_values),
            STATION_ID_DIMNAME: (STATION_ID_DIMNAME, station_ids_strings),
            TIME_DIMNAME: (TIME_DIMNAME, time_coords),
        },
        attrs=x.attrs,
    )
    # Copy data variables from the renamed dataset
    for var_name in x.data_vars:
        if var_name not in (STATION_ID_VARNAME, STATION_NAME_VARNAME):
            # Get the variable from the original dataset
            orig_var = x[var_name]
            # Determine the dimensions for the new variable
            new_dims = []
            for dim in orig_var.dims:
                if dim == ENS_MEMBER_DIMNAME:
                    new_dims.append(REALISATION_DIMNAME)
                elif dim == STATION_DIMNAME:
                    new_dims.append(STATION_ID_DIMNAME)
                else:
                    new_dims.append(dim)
            # Create a new DataArray with the correct dimensions
            new_dataset[var_name] = xr.DataArray(
                data=orig_var.values,
                dims=new_dims,
                coords={dim: new_dataset[dim] for dim in new_dims if dim in new_dataset.coords},
                attrs=orig_var.attrs,
            )
    # Handle station names separately
    station_names_var = x[STATION_NAME_VARNAME]
    new_dataset[STATION_NAME_VARNAME] = xr.DataArray(
        data=station_names_var.values.astype(np.str_),
        dims=[STATION_ID_DIMNAME],
        coords={STATION_ID_DIMNAME: new_dataset[STATION_ID_DIMNAME]},
        attrs=station_names_var.attrs,
    )
    return new_dataset


def template_variable_attributes() -> dict[str, Any]:
    """Return a template dictionary for variable attributes."""
    from efts_io.conventions import _template_variable_attributes

    return _template_variable_attributes()


class EftsDataSet:
    """Convenience class for access to a Ensemble Forecast Time Series in netCDF file."""

    # Reference class convenient for access to a Ensemble Forecast Time Series in netCDF file.
    # Description
    # Reference class convenient for access to a Ensemble Forecast Time Series in netCDF file.

    # Fields
    # time_dim
    # a cached POSIXct vector, the values for the time dimension of the data set.

    # time_zone
    # the time zone for the time dimensions of this data set.

    # identifiers_dimensions
    # a cache, list of values of the primary data identifiers; e.g. station_name or station_id

    # stations_varname
    # name of the variable that stores the names of the stations for this data set.

    def __init__(self, data: Union[str, xr.Dataset]) -> None:
        """Create a new EftsDataSet object."""
        self.time_dim = None
        self.time_zone = "UTC"
        self.time_zone_timestamps = True  # Not sure about https://github.com/csiro-hydroinformatics/efts-io/issues/3
        self.STATION_DIMNAME = STATION_DIMNAME
        self.stations_varname = STATION_ID_VARNAME
        self.LEAD_TIME_DIMNAME = LEAD_TIME_DIMNAME
        self.ENS_MEMBER_DIMNAME = ENS_MEMBER_DIMNAME
        # self.identifiers_dimensions: list = []
        self.data: xr.Dataset
        from pathlib import Path

        if data is None:
            raise ValueError("input cannot be None")
        if isinstance(data, Path):
            data = str(data)
        if isinstance(data, str):
            new_dataset = load_from_stf2_file(data, self.time_zone_timestamps)
            self.data = new_dataset
        elif isinstance(data, xr.Dataset):
            self.data = data
        else:
            raise TypeError(f"Unsupported type {type(data)}")

        self.stf2_int_datatype = "i4"  # default integer type for STF2 saving

    @property
    def title(self) -> str:
        """Get or set the title attribute of the dataset."""
        return self.data.attrs.get(TITLE_ATTR_KEY, "")

    @title.setter
    def title(self, value: str) -> None:
        """Get or set the title attribute of the dataset."""
        self.data.attrs[TITLE_ATTR_KEY] = value

    @property
    def institution(self) -> str:
        """Get or set the institution attribute of the dataset."""
        return self.data.attrs.get(INSTITUTION_ATTR_KEY, "")

    @institution.setter
    def institution(self, value: str) -> None:
        """Get or set the institution attribute of the dataset."""
        self.data.attrs[INSTITUTION_ATTR_KEY] = value

    @property
    def source(self) -> str:
        """Get or set the source attribute of the dataset."""
        return self.data.attrs.get(SOURCE_ATTR_KEY, "")

    @source.setter
    def source(self, value: str) -> None:
        """Get or set the source attribute of the dataset."""
        self.data.attrs[SOURCE_ATTR_KEY] = value

    @property
    def catchment(self) -> str:
        """Get or set the catchment attribute of the dataset."""
        return self.data.attrs.get(CATCHMENT_ATTR_KEY, "")

    @catchment.setter
    def catchment(self, value: str) -> None:
        """Get or set the catchment attribute of the dataset."""
        self.data.attrs[CATCHMENT_ATTR_KEY] = value

    @property
    def stf_convention_version(self) -> float:
        """Get or set the STF_convention_version attribute of the dataset."""
        return self.data.attrs.get(STF_CONVENTION_VERSION_ATTR_KEY, "")

    @stf_convention_version.setter
    def stf_convention_version(self, value: float) -> None:
        """Get or set the STF_convention_version attribute of the dataset."""
        self.data.attrs[STF_CONVENTION_VERSION_ATTR_KEY] = value

    @property
    def stf_nc_spec(self) -> str:
        """Get or set the STF_nc_spec attribute of the dataset."""
        return self.data.attrs.get(STF_NC_SPEC_ATTR_KEY, "")

    @stf_nc_spec.setter
    def stf_nc_spec(self, value: str) -> None:
        """Get or set the STF_nc_spec attribute of the dataset."""
        self.data.attrs[STF_NC_SPEC_ATTR_KEY] = value

    @property
    def comment(self) -> str:
        """Get or set the comment attribute of the dataset."""
        return self.data.attrs.get(COMMENT_ATTR_KEY, "")

    @comment.setter
    def comment(self, value: str) -> None:
        """Get or set the comment attribute of the dataset."""
        self.data.attrs[COMMENT_ATTR_KEY] = value

    @property
    def history(self) -> str:
        """Gets/sets the history attribute of the dataset."""
        return self.data.attrs.get(HISTORY_ATTR_KEY, "")

    @history.setter
    def history(self, value: str) -> None:
        """Gets/sets the history attribute of the dataset."""
        self.data.attrs[HISTORY_ATTR_KEY] = value

    def append_history(self, message: str, timestamp: Optional[datetime] = None) -> None:
        """Append a new entry to the `history` attribute with a timestamp.

        message: The message to append.
        timestamp: If not provided, the current UTC time is used.
        """
        from datetime import UTC

        if timestamp is None:
            timestamp = datetime.now(UTC).isoformat()

        current_history = self.data.attrs.get(HISTORY_ATTR_KEY, "")
        if current_history:
            self.data.attrs[HISTORY_ATTR_KEY] = f"{current_history}\n{timestamp} - {message}"
        else:
            self.data.attrs[HISTORY_ATTR_KEY] = f"{timestamp} - {message}"

    def to_netcdf(self, path: str, version: Optional[str] = "2.0") -> None:
        """Write the data set to a netCDF file."""
        if version is None:
            self.data.to_netcdf(path)
        elif version == "2.0":
            self.save_to_stf2(path)
        else:
            raise ValueError("Only version 2.0 is supported for now")

    def set_mandatory_global_attributes(
        self,
        title: str = "not provided",
        institution: str = "not provided",
        catchment: str = "not provided",
        source: str = "not provided",
        comment: str = "not provided",
        history: str = "not provided",
        append_history: bool = False,  # noqa: FBT001, FBT002
    ) -> None:
        """Sets mandatory global attributes for an EFTS dataset."""
        self.title = title
        self.institution = institution
        self.catchment = catchment
        self.source = source
        self.comment = comment
        if append_history:
            self.append_history(history)
        else:
            self.history = history
        self.stf_convention_version = "2.0"
        self.stf_nc_spec = STF_2_0_URL

    def writeable_to_stf2(self) -> bool:
        """Check if the dataset can be written to a netCDF file compliant with STF 2.0 specification.

        This method checks if the underlying xarray dataset or dataarray has the required dimensions and global attributes as specified by the STF 2.0 convention.

        Returns:
            bool: True if the dataset can be written to a STF 2.0 compliant netCDF file, False otherwise.
        """
        from efts_io.conventions import exportable_to_stf2

        return exportable_to_stf2(self.data)

    @property
    def stf2_int_datatype(self) -> str:
        """The type of integer to save to in the STF 2.x netcdf convention: 'i4' or 'i8'."""
        return self._stf2_int_datatype

    @stf2_int_datatype.setter
    def stf2_int_datatype(self, value: str) -> None:
        """The type of integer to save to in the STF 2.x netcdf convention: 'i4' or 'i8'."""
        if value not in ("i4", "i8"):
            raise ValueError("stf2_int_datatype must be either 'i4' or 'i8'")
        self._stf2_int_datatype = value

    def save_to_stf2(
        self,
        path: str,
        variable_name: Optional[str] = None,
        var_type: StfVariable = StfVariable.STREAMFLOW,
        data_type: StfDataType = StfDataType.OBSERVED,
        ens: bool = False,  # noqa: FBT001, FBT002
        timestep: str = "days",
        data_qual: Optional[xr.DataArray] = None,
    ) -> None:
        """Save to file."""
        from efts_io._ncdf_stf2 import write_nc_stf2

        if isinstance(self.data, xr.Dataset):
            if variable_name is None:
                raise ValueError("Inner data is a DataSet, so an explicit variable name must be explicitely specified.")
            d = self.data[variable_name]
        # elif isinstance(self.data, xr.DataArray):
        #    d = self.data
        else:
            raise TypeError(f"Unsupported data type {type(self.data)}")

        if UNITS_ATTR_KEY not in d.attrs:
            raise ValueError(f"DataArray variable '{d.name}' must have '{UNITS_ATTR_KEY}' attribute defined.")

        write_nc_stf2(
            out_nc_file=path,  # : str,
            dataset=self.data,
            data=d,  # : xr.DataArray,
            var_type=var_type,  # : int = 1,
            data_type=data_type,  # : int = 3,
            stf_nc_vers=2,  # : int = 2,
            ens=ens,  # : bool = False,
            timestep=timestep,  # :str="days",
            data_qual=data_qual,  # : Optional[xr.DataArray] = None,
            overwrite=True,  # :bool=True,
            # loc_info=loc_info, # : Optional[Dict[str, Any]] = None,
            intdata_type=self.stf2_int_datatype,
        )

    def create_data_variables(self, data_var_def: Dict[str, Dict[str, Any]]) -> None:
        """Create data variables in the data set.

        var_defs_dict["variable_1"].keys()
        dict_keys(['name', 'longname', 'units', 'dim_type', 'missval', 'precision', 'attributes'])
        """
        ens_fcast_data_var_def = [x for x in data_var_def.values() if x["dim_type"] == "4"]
        ens_data_var_def = [x for x in data_var_def.values() if x["dim_type"] == "3"]
        point_data_var_def = [x for x in data_var_def.values() if x["dim_type"] == "2"]

        four_dims_names = (LEAD_TIME_DIMNAME, STATION_ID_DIMNAME, REALISATION_DIMNAME, TIME_DIMNAME)
        three_dims_names = (STATION_ID_DIMNAME, REALISATION_DIMNAME, TIME_DIMNAME)
        two_dims_names = (STATION_ID_DIMNAME, TIME_DIMNAME)

        four_dims_shape = tuple(self.data.sizes[dimname] for dimname in four_dims_names)
        three_dims_shape = tuple(self.data.sizes[dimname] for dimname in three_dims_names)
        two_dims_shape = tuple(self.data.sizes[dimname] for dimname in two_dims_names)
        for vardefs, dims_shape, dims_names in [
            (ens_fcast_data_var_def, four_dims_shape, four_dims_names),
            (ens_data_var_def, three_dims_shape, three_dims_names),
            (point_data_var_def, two_dims_shape, two_dims_names),
        ]:
            for x in vardefs:
                varname = x["name"]
                # TODO:
                # _check_mandatory_keys(x)
                self._new_variable_from_legacy_specs(dims_shape, dims_names, x, varname)

    def _new_variable_from_legacy_specs(
        self, dim_shape: Tuple, dims_names: Iterable[str], x: dict[str, Any], varname: str,
    ) -> xr.DataArray:
        """Create a new variable in the data set."""
        data_coords = {dim: self.data.coords[dim] for dim in dims_names}
        new_array = xr.DataArray(
            name=varname,
            data=nan_full(dim_shape),
            coords=data_coords,
            dims=dims_names,
            attrs={
                LONG_NAME_ATTR_KEY: x["longname"],
                UNITS_ATTR_KEY: x[UNITS_ATTR_KEY],
                FILLVALUE_ATTR_KEY: x["missval"],
                "precision": x["precision"],  # TODO: check whether this is still of use.
                **x["attributes"],
            },
        )
        self.data[varname] = new_array
        return new_array

    def new_variable(
        self, varname: str, dim_names: Iterable[str], var_attributes: dict[str, Any], data: Optional[np.ndarray] = None,
    ) -> xr.DataArray:
        """Create a new variable in the data set.

        Args:
            varname (str): Name of the new variable.
            dim_names (Iterable[str]): Names of the dimensions for the new variable.
            var_attributes (dict[str, Any]): Attributes for the new variable. Must include 'units' key. See `template_variable_attributes`
            data (Optional[np.ndarray], optional): Data for the new variable. If None, the variable is initialized with NaNs. Defaults to None.

        Returns:
            xr.DataArray: The newly created variable as an xarray DataArray.
        """
        if varname in self.data.variables:
            raise ValueError(f"Variable '{varname}' already exists in the dataset.")
        if UNITS_ATTR_KEY not in var_attributes:
            raise ValueError(f"Variable attributes must include '{UNITS_ATTR_KEY}' key.")
        known_dimnames = self.get_dim_names()
        unknown_dims = [x for x in dim_names if x not in set(known_dimnames)]
        if unknown_dims:
            raise ValueError(f"Unknown dimension names: {unknown_dims}; must be one of {known_dimnames}.")
        dims_shape = tuple(self.data.sizes[dimname] for dimname in dim_names)
        if data is not None:
            if data.shape != dims_shape:
                raise ValueError(
                    f"Data shape {data.shape} does not match expected shape {dims_shape} for dimensions {dim_names}.",
                )
            data_array = data
        else:
            data_array = nan_full(dims_shape)
        data_coords = {dim: self.data.coords[dim] for dim in dim_names}
        new_array = xr.DataArray(
            name=varname,
            data=data_array,
            coords=data_coords,
            dims=dim_names,
            attrs=var_attributes.copy(),
        )
        self.data[varname] = new_array
        return new_array

    def get_all_series(
        self,
        variable_name: str = "rain_obs",
        dimension_id: Optional[str] = None,  # noqa: ARG002
    ) -> xr.DataArray:
        """Return a multivariate time series, where each column is the series for one of the identifiers."""
        # Return a multivariate time series, where each column is the series for one of the identifiers (self, e.g. rainfall station identifiers):
        return self.data[variable_name]
        # stopifnot(variable_name %in% names(ncfile$var))
        # td = self.get_time_dim()
        # if dimension_id is None: dimension_id = self.get_stations_varname()
        # identifiers = self._get_values(dimension_id)
        # ncdims = self.get_variable_dim_names(variable_name)
        # could be e.g.: double q_obs[lead_time,station,ens_member,time] float
        # rain_obs[station,time] lead_time,station,ens_member,time reordered
        # according to the variable present dimensions:
        # tsstart = splice_named_var(c(1, 1, 1, 1), ncdims)
        # tscount = splice_named_var(c(1, length(identifiers), 1, length(td)), ncdims)
        # rawData = ncdf4::ncvar_get(ncfile, variable_name, start = tsstart, count = tscount,
        # collapse_degen = FALSE)
        # dim_names(rawData) = ncdims
        # # [station,time] to [time, station] for xts creation
        # # NOTE: why can this not be dimension_id instead of STATION_DIMNAME?
        # tsData = reduce_dimensions(rawData,c(TIME_DIMNAME, STATION_DIMNAME))
        # v = xts(x = tsData, order.by = td, tzone = tz(td))
        # colnames(v) = identifiers
        # return(v)

    def get_dim_names(self) -> List[str]:
        """Gets the name of all dimensions in the data set."""
        return [x for x in self.data.sizes.keys()]  # noqa: C416, SIM118
        # Note: self._dim_size will return a list of str in the future
        # return [x for x in self._dim_size.keys()]

    def get_ensemble_for_stations(
        self,
        variable_name: str = "rain_sim",
        identifier: Optional[str] = None,
        dimension_id: str = ENS_MEMBER_DIMNAME,
        start_time: pd.Timestamp = None,
        lead_time_count: Optional[int] = None,
    ) -> xr.DataArray:
        """Not yet implemented."""
        # Return a time series, representing a single ensemble member forecast for all stations over the lead time
        raise NotImplementedError

    def get_ensemble_forecasts(
        self,
        variable_name: str = "rain_sim",
        identifier: Optional[str] = None,
        dimension_id: Optional[str] = None,
        start_time: Optional[pd.Timestamp] = None,
        lead_time_count: Optional[int] = None,
    ) -> xr.DataArray:
        """Gets an ensemble forecast for a variable."""
        # Return a time series, ensemble of forecasts over the lead time
        if dimension_id is None:
            dimension_id = self.get_stations_varname()
        td = self.get_time_dim()
        if start_time is None:
            start_time = td[0]
        n_ens = self.get_ensemble_size()
        raise NotImplementedError(
            "get_ensemble_forecasts: not yet implemented",
        )
        index_id = self.index_for_identifier(identifier, dimension_id)
        check_index_found(index_id, identifier, dimension_id)
        if lead_time_count is None:
            lead_time_count = self.get_lead_time_count()
        indx_time = self.index_for_time(start_time)
        # float rain_sim[lead_time,station,ens_member,time]
        ens_data = self.data.get(variable_name)[
            indx_time,
            :n_ens,
            index_id,
            :lead_time_count,
        ]
        # ensData = self.data.get(variable_name), start = [1, index_id, 1, indTime],
        #     count = c(lead_time_count, 1, nEns, 1), collapse_degen = FALSE)
        # tu = self.get_lead_time_unit()
        # if tu == "days":
        #     timeAxis = start_time + pd.Timedelta(ncfile$dim$lead_time$vals)
        # } else {
        # timeAxis = start_time + lubridate::dhours(1) * ncfile$dim$lead_time$vals
        # }
        # out = xts(x = ensData[, 1, , 1], order.by = timeAxis, tzone = tz(start_time))
        return ens_data  # noqa: RET504

    # def get_ensemble_forecasts_for_station(
    #     self,
    #     variable_name: str = "rain_sim",
    #     identifier: Optional[str] = None,
    #     dimension_id: Optional[str] = None,
    # ):
    #     """Return an array, representing all ensemble member forecasts for a single stations over all lead times."""
    #     if dimension_id is None:
    #         dimension_id = self.get_stations_varname()
    #     raise NotImplementedError

    # def get_ensemble_series(
    #     self,
    #     variable_name: str = "rain_ens",
    #     identifier: Optional[str] = None,
    #     dimension_id: Optional[str] = None,
    # ):
    #     """Return an ensemble of point time series for a station identifier."""
    #     # Return an ensemble of point time series for a station identifier
    #     if dimension_id is None:
    #         dimension_id = self.get_stations_varname()
    #     raise NotImplementedError

    def _dim_size(self, dimname: str) -> int:
        return self.data.sizes[dimname]

    def get_ensemble_size(self) -> int:
        """Return the length of the ensemble size dimension."""
        return self._dim_size(self.ENS_MEMBER_DIMNAME)

    def get_lead_time_count(self) -> int:
        """Length of the lead time dimension."""
        return self._dim_size(self.LEAD_TIME_DIMNAME)

    def get_lead_time_values(self) -> np.ndarray:
        """Return the values of the lead time dimension."""
        return self.data[self.LEAD_TIME_DIMNAME].values

    def put_lead_time_values(self, values: Iterable[float]) -> None:
        """Set the values of the lead time dimension."""
        self.data[self.LEAD_TIME_DIMNAME].values = np.array(values)

    def get_single_series(
        self,
        variable_name: str = "rain_obs",
        identifier: Optional[str] = None,
        dimension_id: Optional[str] = None,
    ) -> xr.DataArray:
        """Return a single point time series for a station identifier."""
        # Return a single point time series for a station identifier. Falls back on def get_all_series if the argument "identifier" is missing
        if dimension_id is None:
            dimension_id = self.get_stations_varname()
        return self.data[variable_name].sel({dimension_id: identifier})

    def get_station_count(self) -> int:
        """Return the number of stations in the data set."""
        self._dim_size(self.STATION_DIMNAME)

    def get_stations_varname(self) -> str:
        """Return the name of the variable that has the station identifiers."""
        # Gets the name of the variable that has the station identifiers
        # TODO: station is integer normally in STF (Euargh)
        return STATION_ID_VARNAME

    def get_time_dim(self) -> np.ndarray:
        """Return the time dimension variable as a vector of date-time stamps."""
        # Gets the time dimension variable as a vector of date-time stamps
        return self.data.time.values  # but loosing attributes.

    # def get_time_unit(self) -> str:
    #     """Return the time units of a read time series."""
    #     # Gets the time units of a read time series, i.e. "hours since 2015-10-04 00:00:00 +1030". Returns the string "hours"
    #     return "dummy"

    # def get_time_zone(self) -> str:
    #     # Gets the time zone to use for the read time series
    #     return "dummy"

    # def get_utc_offset(self, as_string: bool = True):
    #     # Gets the time zone to use for the read time series, i.e. "hours since 2015-10-04 00:00:00 +1030". Returns the string "+1030" or "-0845" if as_string is TRUE, or a lubridate Duration object if FALSE
    #     return None

    # def _get_values(self, variable_name: str):
    #     # Gets (and cache in memory) all the values in a variable. Should be used only for dimension variables
    #     from efts_io.conventions import conventional_varnames

    #     if variable_name not in conventional_varnames:
    #         raise ValueError(
    #             variable_name + " cannot be directly retrieved. Must be in " + ", ".join(conventional_varnames),
    #         )
    #     return self.data[variable_name].values

    # def get_variable_dim_names(self, variable_name):
    #     # Gets the names of the dimensions that define the geometry of a given variable
    #     return [x for x in self.data[[variable_name]].coords.keys()]

    # def get_variable_names(self):
    #     # Gets the name of all variables in the data set
    #     return [x for x in self.data.variables.keys()]

    # def index_for_identifier(self, identifier, dimension_id=None):
    #     # Gets the index at which an identifier is found in a dimension variable
    #     if dimension_id is None:
    #         dimension_id = self.get_stations_varname()
    #     identValues = self._get_values(dimension_id)
    #     if identifier is None:
    #         raise Exception("Identifier cannot be NA")
    #     return _first_where(identifier == identValues)

    # def index_for_time(self, dateTime):
    #     # Gets the index at which a date-time is found in the main time axis of this data set
    #     return _first_where(self.data.time == dateTime)

    # def put_ensemble_forecasts(
    #     self,
    #     x,
    #     variable_name="rain_sim",
    #     identifier: str = None,
    #     dimension_id=None,
    #     start_time=None,
    # ):
    #     # Puts one or more ensemble forecast into a netCDF file
    #     if dimension_id is None:
    #         dimension_id = self.get_stations_varname()
    #     raise NotImplementedError

    # def put_ensemble_forecasts_for_station(
    #     self,
    #     x,
    #     variable_name="rain_sim",
    #     identifier: str = None,
    #     dimension_id=ENS_MEMBER_DIMNAME,
    #     start_time=None,
    # ):
    #     # Puts a single ensemble member forecasts for all stations into a netCDF file
    #     raise NotImplementedError

    # def put_ensemble_series(
    #     self,
    #     x,
    #     variable_name="rain_ens",
    #     identifier: str = None,
    #     dimension_id=None,
    # ):
    #     # Puts an ensemble of time series, e.g. replicate rainfall series
    #     if dimension_id is None:
    #         dimension_id = self.get_stations_varname()
    #     raise NotImplementedError

    # def put_single_series(
    #     self,
    #     x,
    #     variable_name="rain_obs",
    #     identifier: str = None,
    #     dimension_id=None,
    #     start_time=None,
    # ):
    #     # Puts a time series, or part thereof
    #     if dimension_id is None:
    #         dimension_id = self.get_stations_varname()
    #     raise NotImplementedError

    # def put_values(self, x, variable_name):
    #     # Puts all the values in a variable. Should be used only for dimension variables
    #     raise NotImplementedError

    # def set_time_zone(self, tzone_id):
    #     # Sets the time zone to use for the read time series
    #     raise NotImplementedError

    # def summary(self):
    #     # Print a summary of this EFTS netCDF file
    #     raise NotImplementedError

    # See Also
    # See create_efts and open_efts for examples on how to read or write EFTS netCDF files using this dataset.


#' Creates a EftsDataSet for access to a netCDF EFTS data set
#'
#' Creates a EftsDataSet for access to a netCDF EFTS data set
#'
#' @param ncfile name of the netCDF file, or an object of class 'ncdf4'
#' @param writein if TRUE the data set is opened in write mode
#' @export
#' @import ncdf4
#' @examples
#' library(efts)
#' ext_data = system.file('extdata', package='efts')
#' ens_fcast_file = file.path(ext_data, 'Upper_Murray_sample_ensemble_rain_fcast.nc')
#' stopifnot(file.exists(ens_fcast_file))
#' snc = open_efts(ens_fcast_file)
#' (variable_names = snc$get_variable_names())
#' (stations_ids = snc$get_values(STATION_ID_DIMNAME))
#' nEns = snc$get_ensemble_size()
#' nLead = snc$get_lead_time_count()
#' td = snc$get_time_dim()
#' stopifnot('rain_fcast_ens' %in% variable_names)
#'
#' ens_fcast_rainfall = snc$get_ensemble_forecasts('rain_fcast_ens',
#'   stations_ids[1], start_time=td[2])
#' names(ens_fcast_rainfall) = as.character(1:ncol(ens_fcast_rainfall))
#' plot(ens_fcast_rainfall, legend.loc='right')
#'
#' snc$close()
#'
#' @return A EftsDataSet object
#' @importFrom methods is
def open_efts(ncfile: Any, writein: bool = False) -> EftsDataSet:  # noqa: ARG001, FBT001, FBT002
    """Open an EFTS NetCDF file."""
    # raise NotImplemented("open_efts")
    # if isinstance(ncfile, str):
    #     nc = ncdf4::nc_open(ncfile, readunlim = FALSE, write = writein)
    # } else if (methods::is(ncfile, "ncdf4")) {
    #     nc = ncfile
    # }
    return EftsDataSet(ncfile)


def nan_full(shape: Union[Tuple, int]) -> np.ndarray:
    """Create a full array of NaNs with the given shape."""
    if isinstance(shape, int):
        shape = (shape,)
    return np.full(shape=shape, fill_value=np.nan)


def xr_efts(
    issue_times: Iterable[ConvertibleToTimestamp],
    station_ids: Iterable[str],
    lead_times: Optional[Iterable[int]] = None,
    lead_time_tstep: str = "hours",
    ensemble_size: int = 1,
    # variables
    station_names: Optional[Iterable[str]] = None,
    latitudes: Optional[Iterable[float]] = None,
    longitudes: Optional[Iterable[float]] = None,
    areas: Optional[Iterable[float]] = None,
    nc_attributes: Optional[Dict[str, str]] = None,
) -> xr.Dataset:
    """Create an xarray Dataset for EFTS data."""
    # Check that station ids are unique:
    if len(set(station_ids)) != len(station_ids):
        raise ValueError("Station names must be unique.")
    # I learned today that xarray 2025.7.1 can now accept pandas datetimeindex as coordinates
    # for backward compatibility with older xarray versions, we convert to list here.
    # See https://github.com/csiro-hydroinformatics/efts-io/issues/13, in the future may change design.
    if isinstance(issue_times, pd.DatetimeIndex):
        # This will convert each item to a tstamp such as
        # Timestamp('2023-01-01 00:00:00+1000', tz='UTC+10:00')
        issue_times = list(issue_times)  # issue_times is iterable,and iterated over indeed.
    if lead_times is None:
        lead_times = [0]
    coords = {
        TIME_DIMNAME: issue_times,
        # STATION_DIMNAME: np.arange(start=1, stop=len(station_ids) + 1, step=1),
        STATION_ID_DIMNAME: station_ids,  # np.arange(start=1, stop=len(station_ids) + 1, step=1),
        REALISATION_DIMNAME: np.arange(start=1, stop=ensemble_size + 1, step=1),
        LEAD_TIME_DIMNAME: lead_times,
        # Initially, I was exploring attaching a coordinate to an existing dimension STATION_DIMNAME, using:
        # https://docs.xarray.dev/en/latest/generated/xarray.DataArray.assign_coords.html#xarray.DataArray.assign_coords
        # then using https://github.com/pydata/xarray/issues/2028#issuecomment-1265252754  to be able to
        # index by station IDs. But in July 2025 decided to not have a STATION_DIMNAME dimension, which is
        # an artefact from legacy conventions (Fortran 1-based indexing and other related limitations).
        # Keeping a number based STATION_DIMNAME here is only making things more difficult and data subsetting more prone to bugs.
        # STATION_ID_VARNAME: (STATION_DIMNAME, station_ids),
    }
    n_stations = len(station_ids)
    latitudes = latitudes if latitudes is not None else nan_full(n_stations)
    longitudes = longitudes if longitudes is not None else nan_full(n_stations)
    areas = areas if areas is not None else nan_full(n_stations)
    station_names = station_names if station_names is not None else [f"{i}" for i in station_ids]
    data_vars = {
        STATION_NAME_VARNAME: (STATION_ID_DIMNAME, station_names),
        LAT_VARNAME: (STATION_ID_DIMNAME, latitudes),
        LON_VARNAME: (STATION_ID_DIMNAME, longitudes),
        AREA_VARNAME: (STATION_ID_DIMNAME, areas),
    }
    nc_attributes = nc_attributes or _stf2_mandatory_global_attributes()
    d = xr.Dataset(
        data_vars=data_vars,
        coords=coords,
        attrs=nc_attributes,
    )
    # Credits to the work reported in https://github.com/pydata/xarray/issues/2028#issuecomment-1265252754
    # d = d.set_xindex(STATION_ID_VARNAME)
    d.time.attrs = {
        STANDARD_NAME_ATTR_KEY: TIME_DIMNAME,
        LONG_NAME_ATTR_KEY: TIME_DIMNAME,
        # TIME_STANDARD_KEY: "UTC",
        AXIS_ATTR_KEY: "t",
        # UNITS_ATTR_KEY: "days since 2000-11-14 23:00:00.0 +0000",
    }
    d.lead_time.attrs = {
        STANDARD_NAME_ATTR_KEY: "lead time",
        LONG_NAME_ATTR_KEY: "forecast lead time",
        AXIS_ATTR_KEY: "v",
        UNITS_ATTR_KEY: f"{lead_time_tstep} since time",
    }
    d.realisation.attrs = {
        STANDARD_NAME_ATTR_KEY: ENS_MEMBER_DIMNAME,  # TODO: should we keep the STF 2.0 ens_member as a standard name?
        LONG_NAME_ATTR_KEY: "ensemble member",
        UNITS_ATTR_KEY: "member id",
        AXIS_ATTR_KEY: "u",
    }
    d.station_id.attrs = {LONG_NAME_ATTR_KEY: "station or node identification code"}
    d.station_name.attrs = {LONG_NAME_ATTR_KEY: "station or node name"}
    d.lat.attrs = {LONG_NAME_ATTR_KEY: "latitude", UNITS_ATTR_KEY: "degrees_north", AXIS_ATTR_KEY: "y"}
    d.lon.attrs = {LONG_NAME_ATTR_KEY: "longitude", UNITS_ATTR_KEY: "degrees_east", AXIS_ATTR_KEY: "x"}
    d.area.attrs = {
        LONG_NAME_ATTR_KEY: "station area",
        UNITS_ATTR_KEY: "km^2",
        STANDARD_NAME_ATTR_KEY: AREA_VARNAME,
    }
    return d


def create_mandatory_global_attributes(
    title: str,
    institution: str,
    catchment: str,
    source: str,
    comment: str,
    history: Optional[str] = None,
) -> Dict[str, str]:
    """Create a dictionary of mandatory global attributes for an EFTS dataset.

    Args:
        title (str): Title of the dataset.
        institution (str): Institution responsible for the dataset.
        catchment (str): Catchment area description.
        source (str): Source of the data.
        comment (str): Additional comments about the dataset.
        history (Optional[str], optional): History of the dataset. If None, a default history message is created. Defaults to None.

    Returns:
        Dict[str, str]: A dictionary containing the mandatory global attributes.
    """
    d = _stf2_mandatory_global_attributes(
        title=title,
        institution=institution,
        catchment=catchment,
        source=source,
        comment=comment,
        history=history or f"Created on {pd.Timestamp.now(tz='UTC').isoformat()}",
    )
    return d  # noqa: RET504


def _stf2_mandatory_global_attributes(
    title: str = "not provided",
    institution: str = "not provided",
    catchment: str = "not provided",
    source: str = "not provided",
    comment: str = "not provided",
    history: str = "not provided",
) -> Dict[str, str]:
    """Create a dictionary of mandatory global attributes for an EFTS dataset."""
    return {
        TITLE_ATTR_KEY: title,
        INSTITUTION_ATTR_KEY: institution,
        CATCHMENT_ATTR_KEY: catchment,
        SOURCE_ATTR_KEY: source,
        COMMENT_ATTR_KEY: comment,
        HISTORY_ATTR_KEY: history,
        STF_CONVENTION_VERSION_ATTR_KEY: "2.0",
        STF_NC_SPEC_ATTR_KEY: STF_2_0_URL,
    }


#' Creates a EftsDataSet for write access to a netCDF EFTS data set
#'
#' Creates a EftsDataSet for write access to a netCDF EFTS data set
#'
#' @param fname file name to create to. The file must not exist already.
#' @param time_dim_info a list with the units and values defining the time dimension of the data set
#' @param data_var_definitions a data frame, acceptable by \code{\link{create_variable_definitions}}, or list of netCDF variable definitions, e.g.
#'       \code{list(rain_sim=list(name='rain_sim', longname='ECMWF Rainfall ensemble forecasts', units='mm', missval=-9999.0, precision='double', attributes=list(type=2, type_description='accumulated over the preceding interval')))}
#' @param stations_ids station identifiers, coercible to an integer vector (note: may change to be a more flexible character storage)
#' @param station_names optional; names of the stations
#' @param nc_attributes a named list of characters, attributes for the whole file,
#' including mandatory ones: title, institution, source, catchment, comment.
#' You may use \code{\link{create_global_attributes}} as a starting template.
#' @param lead_length length of the lead forecasting time series.
#' @param optional_vars a data frame defining optional netCDF variables. For a templated default see
#' \code{\link{default_optional_variable_definitions_v2_0}} and
#' \url{https://github.com/jmp75/efts/blob/107c553045a37e6ef36b2eababf6a299e7883d50/docs/netcdf_for_water_forecasting.md#optional-variables}
#' @param lead_time_tstep string specifying the time step of the forecast lead length.
#' @param ensemble_length number of ensembles, i.e. number of forecasts for each point on the main time axis of the data set
#' @examples
#'
#' # NOTE
#' # The sample code below is purposely generic; to produce
#' # a data set conforming with the conventions devised for
#' # ensemble streamflow forecast you will need to
#' # follow the additional guidelines at
#' # https://github.com/jmp75/efts/blob/master/docs/netcdf_for_water_forecasting.md
#'
#' fname = tempfile()
#'
#' stations_ids = c(123,456)
#' nEns = 3
#' nLead = 4
#' nTimeSteps = 12
#'
#' timeAxisStart = ISOdate(year=2010, month=08, day=01, hour = 14, min = 0, sec = 0, tz = 'UTC')
#' time_dim_info = create_time_info(from=timeAxisStart,
#'   n=nTimeSteps, time_step = "hours since")
#'
#' # It is possible to define variables for three combinations of dimensions.
#' # dimensions '4' ==> [lead_time,station,ens_member,time]
#' # dimensions '3' ==> [station,ens_member,time]
#' # dimensions '2' ==> [station,time]
#'
#' variable_names = c('var1_fcast_ens','var2_fcast_ens', 'var1_obs',
#'   'var2_obs', 'var1_ens','var2_ens')
#'
#' va = create_var_attribute_definition(
#'   type = 2L,
#'   type_description = "accumulated over the preceding interval",
#'   dat_type = "der",
#'   dat_type_description = paste(rep(c("var1", "var2"), 3), "synthetic test data"),
#'   location_type = "Point")
#'
#'
#' (varDef = create_variable_definition_dataframe(
#'   variable_names=variable_names,
#'   long_names = paste(variable_names, 'synthetic data'),
#'   dimensions = c(4L,4L,2L,2L,3L,3L),
#'   var_attributes = va))
#'
#' glob_attr = create_global_attributes(
#'   title="data set title",
#'   institution="my org",
#'   catchment="Upper_Murray",
#'   source="A journal reference, URL",
#'   comment="example for vignette")
#'
#' (opt_metadatavars = default_optional_variable_definitions_v2_0())
#'
#' snc = create_efts(
#'   fname=fname,
#'   time_dim_info=time_dim_info,
#'   data_var_definitions=varDef,
#'   stations_ids=stations_ids,
#'   nc_attributes=glob_attr,
#'   optional_vars = opt_metadatavars,
#'   lead_length=nLead,
#'   ensemble_length=nEns,
#'   lead_time_tstep = "hours")
#'
#' # Following is code that was used to create unit tests for EFTS.
#' # This is kept in this example to provide sample on now to write data of various dimension.
#' td = snc$get_time_dim()
#' m = matrix(ncol=nEns, nrow=nLead)
#' for (rnum in 1:nLead) {
#'     for (cnum in 1:nEns) {
#'       m[rnum,cnum] = rnum*0.01 + cnum*0.1
#'   }
#' }
#' #      [,1] [,2] [,3]
#' # [1,] 0.11 0.21 0.31
#' # [2,] 0.12 0.22 0.32
#' # [3,] 0.13 0.23 0.33
#' # [4,] 0.14 0.24 0.34
#' for (i in 1:length(td)) {
#'   for (j in 1:length(stations_ids)) {
#'     station = stations_ids[j]
#'     var1Values = i + 0.1*j + m
#'     var2Values = 2*var1Values
#'     dtime = td[i]
#'     snc$put_ensemble_forecasts(var1Values, variable_name = variable_names[1],
#'       identifier = station, start_time = dtime)
#'     snc$put_ensemble_forecasts(var2Values, variable_name = variable_names[2],
#'       identifier = station, start_time = dtime)
#'   }
#' }
#'
#' timeSteps = 1:length(td)
#' for (j in 1:length(stations_ids)) {
#'   var3Values = timeSteps + 0.1*j
#'   var4Values = var3Values + 0.01*timeSteps + 0.001*j
#'
#'   station = stations_ids[j]
#'   snc$put_single_series(var3Values, variable_name = variable_names[3], identifier = station)
#'   snc$put_single_series(var4Values, variable_name = variable_names[4], identifier = station)
#' }
#'
#' for (j in 1:length(stations_ids)) {
#'
#'   var5Xts = matrix(rep(1:nEns, each=nTimeSteps) + timeSteps + 0.1*j, ncol=nEns)
#'
#'   # [time,ens_member] to [ens_member,time], as expected by put_ensemble_series
#'   var5Values = t(var5Xts)
#'   var6Values = 0.25 * var5Values
#'
#'   station = stations_ids[j]
#'   snc$put_ensemble_series(var5Values, variable_name = variable_names[5], identifier = station)
#'   snc$put_ensemble_series(var6Values, variable_name = variable_names[6], identifier = station)
#' }
#'
#' # We can get/put values for some metadata variables:
#' snc$get_values("x")
#' snc$put_values(c(1.1, 2.2), "x")
#' snc$put_values(letters[1:2], STATION_NAME_VARNAME)
#'
#' # Direct get/set access to data variables, however, is prevented;
#' #  the following would thus cause an error:
#' # snc$get_values("var1_fcast_ens")
#'
#' snc$close()
#' # Cleaning up temp file:
#' if (file.exists(fname))
#'   file.remove(fname)
#'
#'
#'
#' @export
#' @import ncdf4
#' @importFrom utils packageDescription
#' @importFrom methods new
#' @return A EftsDataSet object
def create_efts(
    fname: str,
    time_dim_info: Dict,
    data_var_definitions: List[Dict[str, Any]],
    stations_ids: List[int],
    station_names: Optional[List[str]] = None,  # noqa: ARG001
    nc_attributes: Optional[Dict[str, str]] = None,
    optional_vars: Optional[dict[str, Any]] = None,
    lead_length: int = 48,
    ensemble_length: int = 50,
    lead_time_tstep: str = "hours",
) -> EftsDataSet:
    """Create a new EFTS dataset."""
    import xarray as xr

    from efts_io.conventions import mandatory_global_attributes

    if stations_ids is None:
        raise ValueError(
            "You must provide station identifiers when creating a new EFTS netCDF data set",
        )

    if nc_attributes is None:
        raise ValueError(
            "You must provide a suitable list for nc_attributes, including" + ", ".join(mandatory_global_attributes),
        )

    # check_global_attributes(nc_attributes)

    if os.path.exists(fname):
        raise FileExistsError("File already exists: " + fname)

    if isinstance(data_var_definitions, pd.DataFrame):
        raise TypeError(
            "data_var_definitions should be a list of dictionaries, not a pandas DataFrame",
        )

    var_defs = create_efts_variables(
        data_var_definitions,
        time_dim_info,
        num_stations=len(stations_ids),
        lead_length=lead_length,
        ensemble_length=ensemble_length,
        optional_vars=optional_vars,
        lead_time_tstep=lead_time_tstep,
    )

    ## attributes for dimensions variables
    def add_dim_attribute(v: xr.Variable, dimname: str, attr_key: str, attr_value: str) -> None:
        pass

    add_dim_attribute(var_defs, TIME_DIMNAME, STANDARD_NAME_ATTR_KEY, TIME_DIMNAME)
    add_dim_attribute(var_defs, TIME_DIMNAME, TIME_STANDARD_ATTR_KEY, "UTC")
    add_dim_attribute(var_defs, TIME_DIMNAME, AXIS_ATTR_KEY, "t")
    add_dim_attribute(var_defs, ENS_MEMBER_DIMNAME, STANDARD_NAME_ATTR_KEY, ENS_MEMBER_DIMNAME)
    add_dim_attribute(var_defs, ENS_MEMBER_DIMNAME, AXIS_ATTR_KEY, "u")
    add_dim_attribute(var_defs, LEAD_TIME_DIMNAME, STANDARD_NAME_ATTR_KEY, LEAD_TIME_DIMNAME)
    add_dim_attribute(var_defs, LEAD_TIME_DIMNAME, AXIS_ATTR_KEY, "v")
    add_dim_attribute(var_defs, LAT_VARNAME, AXIS_ATTR_KEY, "y")
    add_dim_attribute(var_defs, LON_VARNAME, AXIS_ATTR_KEY, "x")

    d = xr.Dataset(
        data_vars=var_defs["datavars"],
        coords=var_defs["metadatavars"],
        attrs={"description": "TODO: put the right attributes"},
    )

    ## Determine if there is real value in a tryCatch. What is the point if we cannot close/delete the file.
    # nc = tryCatch(
    #   createSchema(fname, varDefs, data_var_definitions, nc_attributes, optional_vars,
    #     stations_ids, lead_length, ensemble_length, station_names),
    #   error = function(e) {
    #     stop(paste("netCDF schema creation failed", e))
    #     None
    #   }, finally = function() {
    #   }
    # )
    # nc = createSchema(fname, varDefs, data_var_definitions, nc_attributes, optional_vars,
    #   stations_ids, lead_length, ensemble_length, station_names)

    return EftsDataSet(d)


# ########################################
# # Below are functions not exported
# ########################################

# infoList(theList) {
#   paste(paste(names(theList), theList, sep = ": "), collapse = ", ")
# }

# createSchema(fname, varDefs, data_var_definitions, nc_attributes, optional_vars,
#   stations_ids, lead_length, ensemble_length, station_names=NA) {

#   allVars = c(varDefs$datavars, varDefs$metadatavars)
#   nc = ncdf4::nc_create(fname, vars = allVars)

#   ## attributes for data variables
#   lapply(data_var_definitions, put_variable_attributes, nc)

#   ## attributes for dimensions variables
#   ncdf4::ncatt_put(nc, TIME_DIMNAME, STANDARD_NAME_KEY, TIME_DIMNAME)
#   ncdf4::ncatt_put(nc, TIME_DIMNAME, TIME_STANDARD_KEY, "UTC")
#   ncdf4::ncatt_put(nc, TIME_DIMNAME, AXIS_ATTR_KEY, "t")
#   ncdf4::ncatt_put(nc, ENS_MEMBER_DIMNAME, STANDARD_NAME_KEY, ENS_MEMBER_DIMNAME)
#   ncdf4::ncatt_put(nc, ENS_MEMBER_DIMNAME, AXIS_ATTR_KEY, "u")
#   ncdf4::ncatt_put(nc, LEAD_TIME_DIMNAME, STANDARD_NAME_KEY, LEAD_TIME_DIMNAME)
#   ncdf4::ncatt_put(nc, LEAD_TIME_DIMNAME, AXIS_ATTR_KEY, "v")
#   ncdf4::ncatt_put(nc, LAT_VARNAME, AXIS_ATTR_KEY, "y")
#   ncdf4::ncatt_put(nc, lon_varname, AXIS_ATTR_KEY, "x")

#   ## attributes for optional metadata variables
#   if(!is.None(optional_vars))
#   {
#     var_names = rownames(optional_vars)
#     if(STANDARD_NAME_KEY %in% colnames(optional_vars)){
#       for (v in var_names) {
#         sn = optional_vars[v, STANDARD_NAME_KEY]
#         if(!is.na(sn)) ncdf4::ncatt_put(nc, v, STANDARD_NAME_KEY, sn)
#       }
#     }
#     if(x_varname %in% var_names){
#       ncdf4::ncatt_put(nc, x_varname, AXIS_ATTR_KEY, "x")
#     }
#     if(y_varname %in% var_names){
#       ncdf4::ncatt_put(nc, y_varname, AXIS_ATTR_KEY, "y")
#     }
#   }

#   ## Add global attributes
#   ncdf4::ncatt_put(nc, 0, STF_CONVENTION_VERSION_ATTR_KEY, 2)
#   ncdf4::ncatt_put(nc, 0, STF_NC_SPEC_ATTR_KEY, "https://github.com/jmp75/efts/blob/107c553045a37e6ef36b2eababf6a299e7883d50/docs/netcdf_for_water_forecasting.md")
#   ncdf4::ncatt_put(nc, 0, HISTORY_ATTR_KEY,
#     paste(
#       as.character(lubridate::now(tzone="UTC")),
#       "UTC",
#       "file created with the R package efts", packageDescription("efts")$Version
#     ) %>% infoList)

#   if(!is.None(nc_attributes)) {
#     for (k in names(nc_attributes)) {
#       pad_global_attribute(nc, k, nc_attributes[k])
#     }
#   }

#   ## populate metadata variables
#   ncdf4::ncvar_put(nc, STATION_ID_VARNAME, stations_ids)
#   ncdf4::ncvar_put(nc, LEAD_TIME_DIMNAME, 1:lead_length)
#   ncdf4::ncvar_put(nc, ENS_MEMBER_DIMNAME, 1:ensemble_length)
#   if (!is.None(station_names)) {
#     ncdf4::ncvar_put(nc, STATION_NAME_VARNAME, station_names)
#   }
#   # One seems to need to close/reopen the newly created file, otherwise some
#   # ncvar_get operations will fail with a cryptic message.  I follow the
#   # advice in this and associated posts
#   # https://www.unidata.ucar.edu/mailing_lists/archives/netcdfgroup/2012/msg00270.html
#   ncdf4::nc_close(nc)
#   nc = ncdf4::nc_open(fname, write = TRUE, readunlim = FALSE)
#   return(nc)
# }



"""Handling of EFTS netCDF variables definitions."""

from typing import Any, Dict, Optional, Tuple

# import netCDF4
import numpy as np
import pandas as pd
import xarray as xr

from efts_io._internals import create_data_variable
from efts_io.attributes import create_var_attribute_definition
from efts_io.conventions import (
    AREA_VARNAME,
    ENS_MEMBER_DIMNAME,
    FILLVALUE_ATTR_KEY,
    LEAD_TIME_DIMNAME,
    STANDARD_NAME_ATTR_KEY,
    STATION_DIMNAME,
    UNITS_ATTR_KEY,
)
from efts_io.dimensions import _create_nc_dims


#' Create a variable definition
#'
#' Create a variable definition usable by the function \code{\link{create_efts_variables}} to create netCDF variables.
#'
#' @param name variable name
#' @param longname variable long name
#' @param units variable units
#' @param missval value code for missing data
#' @param precision precision
#' @param dim_type dimension type (EFTS integer code)
#' @param var_attribute list of attributes for the netCDF variable to create
#' @export
#' @return a list
#' @examples
#' var_def = create_variable_definition(name='rain_der',
#'   longname='Rainfall ensemble forecast derived from some prediction', units='mm',
#'   missval=-9999.0, precision='double', var_attribute=list(type=2L,
#'     description="accumulated over the preceding interval",
#'     dat_type = "der", dat_type_description="AWAP data interpolated from observations",
#'     location_type = "Point"))
def create_variable_definition(
    name: str,
    longname: str = "",
    units: str = "mm",
    missval: float = -9999.0,
    precision: str = "double",
    dim_type: str = "4",
    var_attribute: Optional[dict[str, str]] = None,
) -> dict[str, Any]:
    """Create a variable definition."""
    if var_attribute is None:
        var_attribute = create_var_attribute_definition()
    return {
        "name": name,
        "longname": longname,
        UNITS_ATTR_KEY: units,
        "dim_type": dim_type,
        "missval": missval,
        "precision": precision,
        "attributes": var_attribute,
    }


# #' Create a variables definition data frame
# #'
# #' Create a variable definition usable by the function \code{\link{create_variable_definitions}}
# #' to create netCDF variables. The use of this function is not compulsory to create a EFTS
# #' netCDF schema, just offered as a convenience.
# #'
# #' @param variable_names character vector, names of the variables
# #' @param long_names character vector, long names of the variables (defaults to variable_names if missing)
# #' @param standard_names character vector, standard names of the variables (optional, defaults to variable_names)
# #' @param units character vector, units for the variable(s)
# #' @param missval numeric vector, missing value code(s) for the variable(s)
# #' @param precision character vector, precision of the variables
# #' @param dimensions character or integer vector, number of dimensions each variable (2, 3 or 4)
# #' @param var_attributes a list of named attributes. See \code{\link{create_var_attribute_definition}}
# #' @export
# #' @return a data frame suitable for \code{\link{create_variable_definition}}
# #' @seealso See
# #'    \code{\link{create_variable_definition}} and \code{\link{create_efts}} for examples
# create_variable_definition_dataframe(variable_names, long_names = variable_names, standard_names = variable_names, units = "mm", missval = -9999.0,
#   precision = "double", dimensions = 4L, var_attributes = create_var_attribute_definition()) {
#   stopifnot(is.character(variable_names))
#   varsDef = data.frame(name = variable_names, stringsAsFactors = FALSE)
#   varsDef$longname = long_names
#   varsDef$standard_name = standard_names
#   varsDef$units = units
#   varsDef$missval = missval
#   varsDef$precision = precision
#   varsDef$dimensions = as.integer(dimensions)

#   va = data.frame(var_attributes, stringsAsFactors = FALSE)
#   if(nrow(va) < nrow(varsDef)) {
#     va = va[ rep(1:nrow(va), length.out=nrow(varsDef)), ]
#   }

#   varsDef = cbind(varsDef, va)
#   rownames(varsDef) = varsDef$name
#   return(varsDef)
# }


#' Provide a template definition of optional geolocation variables
#'
#' Provide a template definition of optional geolocation and geographic variables x, y, area and elevation.
#' See \url{https://github.com/jmp75/efts/blob/107c553045a37e6ef36b2eababf6a299e7883d50/docs/netcdf_for_water_forecasting.md#optional-variables}.
#'
#' @export
#' @return a data frame
#' @seealso See
#'    \code{\link{create_variable_definition}} and \code{\link{create_efts}} for examples
#' @export
def default_optional_variable_definitions_v2_0() -> pd.DataFrame:
    """Provide a template definition of optional geolocation variables."""
    return pd.DataFrame.from_dict(
        {
            "name": ["x", "y", AREA_VARNAME, "elevation"],
            "longname": [
                "easting from the GDA94 datum in MGA Zone 55",
                "northing from the GDA94 datum in MGA Zone 55",
                "catchment area",
                "station elevation above sea level",
            ],
            STANDARD_NAME_ATTR_KEY: [
                "northing_GDA94_zone55",
                "easting_GDA94_zone55",
                AREA_VARNAME,
                "elevation",
            ],
            UNITS_ATTR_KEY: ["", "", "km^2", "m"],
            "missval": [np.nan, np.nan, -9999.0, -9999.0],
            "precision": np.repeat("float", 4),
        },
    )


# ########################################
# # Below are functions not exported
# ########################################


#' Create variable definitions from a data frame
#'
#' Given a data frame as input, create a list of variable definitions usable by the function \code{\link{create_efts_variables}} to create netCDF variables.
#'
#' @param dframe a data frame, one line is one variable definition. Must have at least the following column names: 'name', 'longname', 'units', 'missval', 'precision', 'type', 'type_description', 'location_type'
#' @export
#' @return a list of length equal to the number of rows in the input data frame
#' @seealso See
#'    \code{\link{create_efts}} for examples
#' @examples
#' varsDef = data.frame(name=letters[1:3], stringsAsFactors=FALSE)
#' varsDef$longname=paste('long name for', varsDef$name)
#' varsDef$units='mm'
#' varsDef$missval=-999.0
#' varsDef$precision='double'
#' varsDef$type=2
#' varsDef$type_description='accumulated over the previous time step'
#' varsDef$location_type='Point'
#' str(create_variable_definitions(varsDef))
#'
def create_variable_definitions(dframe: pd.DataFrame) -> Dict[str, Any]:
    """Create variable definitions from a data frame."""
    in_names = dframe.columns
    non_opt_attr = ["name", "longname", UNITS_ATTR_KEY, "missval", "precision", "dimensions"]
    varargs_attr = [x for x in in_names if x not in non_opt_attr]

    def dataframe_to_dict(df: pd.DataFrame, columns: list) -> dict:
        """Convert a single-row DataFrame to a dictionary for specified columns."""
        if not isinstance(df, pd.Series):
            raise TypeError("single row of a data frame: expected a pandas series")
        return {col: df[col] for col in columns if col in df}

    def f(var_def: Dict[str, Any]):  # noqa: ANN202
        return create_variable_definition(
            name=var_def["name"],
            longname=var_def["longname"],
            units=var_def[UNITS_ATTR_KEY],
            missval=var_def["missval"],
            precision=var_def["precision"],
            dim_type=var_def["dimensions"],
            var_attribute=dataframe_to_dict(var_def, varargs_attr),
        )

    # dframe[['rownum']] = 1:nrow(dframe)
    # r = plyr::dlply(.data = dframe, .variables = "rownum", .fun = f)
    variables_defs: Dict = dframe.apply(lambda x: f(x), axis=1).to_dict()
    return {v["name"]: v for _, v in variables_defs.items()}


def create_mandatory_vardefs(
    station_dim: Tuple[str, np.ndarray, Dict[str, str]],
    str_dim: Tuple[str, np.ndarray, Dict[str, str]],
    ensemble_dim: Tuple[str, np.ndarray, Dict[str, str]],
    lead_time_dim: Tuple[str, np.ndarray, Dict[str, str]],
    lead_time_tstep: str = "hours",
) -> Dict[str, xr.Variable]:
    """Create mandatory variable definitions."""
    # https://github.com/jmp75/efts/blob/107c553045a37e6ef36b2eababf6a299e7883d50/docs/netcdf_for_water_forecasting.md#mandatory-variables
    # float time(time)
    # int station_id(station)
    # char station_name(strLen, station)
    # int ens_member(ens_member)
    # float lead_time(lead_time)
    # float lat (station)
    # float lon (station)

    # STATION_DIMNAME,
    # LEAD_TIME_DIMNAME,
    # TIME_DIMNAME,
    # ENS_MEMBER_DIMNAME,
    # STR_LEN_DIMNAME,

    station_id_variable = xr.Variable(
        dims=[STATION_DIMNAME],
        data=station_dim[1],
        encoding={FILLVALUE_ATTR_KEY: None},
        attrs={
            "longname": station_dim[2]["longname"],
            UNITS_ATTR_KEY: "",
            "missval": None,
            "precision": "integer",
        },
    )
    station_names_dim_variable = xr.Variable(
        dims=[str_dim[0], STATION_DIMNAME],
        # That was not intuitive to create this empty array. Not entirely sure this is what we want.
        data=np.empty_like(
            prototype=b"",
            shape=(len(str_dim[1]), len(station_dim[1])),
            dtype=np.bytes_,
        ),
        encoding={FILLVALUE_ATTR_KEY: None},
        attrs={
            "longname": "station or node name",
            UNITS_ATTR_KEY: "",
            "missval": None,
            "precision": "char",
        },
    )
    ensemble_member_id_variable = xr.Variable(
        dims=[ENS_MEMBER_DIMNAME],
        data=ensemble_dim[1],
        encoding={FILLVALUE_ATTR_KEY: None},
        attrs={
            "longname": ensemble_dim[2]["longname"],
            UNITS_ATTR_KEY: "",
            "missval": None,
            "precision": "integer",
        },
    )
    lead_time_dim_variable = xr.Variable(
        dims=[LEAD_TIME_DIMNAME],
        data=lead_time_dim[1],
        encoding={FILLVALUE_ATTR_KEY: None},
        attrs={
            "longname": lead_time_dim[2]["longname"],
            UNITS_ATTR_KEY: lead_time_tstep + " since time",
            "missval": None,
            "precision": "integer",
        },
    )
    latitude_dim_variable = xr.Variable(
        dims=[STATION_DIMNAME],
        data=np.empty_like(station_dim[1], dtype=float),
        encoding={FILLVALUE_ATTR_KEY: -9999.0},
        attrs={
            "longname": "latitude",
            UNITS_ATTR_KEY: "degrees north",
            "missval": -9999.0,
            "precision": "float",
        },
    )
    longitude_dim_variable = xr.Variable(
        dims=[STATION_DIMNAME],
        data=np.empty_like(station_dim[1], dtype=float),
        encoding={FILLVALUE_ATTR_KEY: -9999.0},
        attrs={
            "longname": "longitude",
            UNITS_ATTR_KEY: "degrees east",
            "missval": -9999.0,
            "precision": "float",
        },
    )

    return {
        "station_ids_var": station_id_variable,
        "station_names_var": station_names_dim_variable,
        "ensemble_var": ensemble_member_id_variable,
        "lead_time_var": lead_time_dim_variable,
        "latitude_var": latitude_dim_variable,
        "longitude_var": longitude_dim_variable,
    }


def create_optional_vardefs(
    station_dim: Tuple[str, np.ndarray, Dict[str, str]],
    vars_def: Optional[pd.DataFrame] = None,
) -> pd.Series:
    """Create optional variable definitions."""
    if vars_def is None:
        vars_def = default_optional_variable_definitions_v2_0()

    # https://github.com/jmp75/efts/blob/107c553045a37e6ef36b2eababf6a299e7883d50/docs/netcdf_for_water_forecasting.md#mandatory-variables
    # vars_def$rownum = 1:nrow(vars_def)
    def f(vd: Dict):  # noqa: ANN202
        return {
            "name": vd["name"],
            UNITS_ATTR_KEY: vd[UNITS_ATTR_KEY],
            "dim": list(station_dim[0]),  # TOCHECK or not a list but the str?
            "missval": vd["missval"],
            "longname": vd["longname"],
            "prec": vd["precision"],
        }

    return vars_def.apply(lambda x: f(x), axis=1)


#' Create netCDF variables according to the definition
#'
#' Create netCDF variables according to the definition
#'
#' @param data_var_def a list, with each item itself a list suitable as a variable definition argument to create_data_variable
#' @param time_dim_info a list with the units and values defining the time dimension of the data set
#' @param num_stations number of (gauging) stations identifying points in the data set
#' @param lead_length length of the lead forecasting time series.
#' @param ensemble_length number of ensembles, i.e. number of forecasts for each point on the main time axis of the data set
#' @param optional_vars a data frame defining optional netCDF variables. For a templated default see
#' \code{\link{default_optional_variable_definitions_v2_0}} and
#' \url{https://github.com/jmp75/efts/blob/107c553045a37e6ef36b2eababf6a299e7883d50/docs/netcdf_for_water_forecasting.md#optional-variables}
#' @param lead_time_tstep string specifying the time step of the forecast lead length.
#' @seealso See
#'    \code{\link{create_efts}} for examples
def create_efts_variables(
    data_var_def: Dict,
    time_dim_info: Dict,
    num_stations: int,
    lead_length: int,
    ensemble_length: int,
    optional_vars: Optional[pd.DataFrame],
    lead_time_tstep: str,
) -> Dict[str, Any]:
    """Create netCDF variables according to the definition."""
    efts_dims = _create_nc_dims(
        time_dim_info=time_dim_info,
        num_stations=num_stations,
        lead_length=lead_length,
        ensemble_length=ensemble_length,
    )

    time_dim = efts_dims["time_dim"]
    lead_time_dim = efts_dims["lead_time_dim"]
    station_dim = efts_dims["station_dim"]
    str_dim = efts_dims["str_dim"]
    ensemble_dim = efts_dims["ensemble_dim"]

    mandatory_var_ncdefs = create_mandatory_vardefs(
        station_dim,
        str_dim,
        ensemble_dim,
        lead_time_dim,
        lead_time_tstep,
    )
    variables_metadata = mandatory_var_ncdefs
    if optional_vars is not None:
        optional_var_ncdefs = create_optional_vardefs(
            station_dim,
            vars_def=optional_vars,
        )
        # TODO if not native to ncdf4: check name clashes
        # already_defs = names(variables)
        variables_metadata.update(optional_var_ncdefs)

    unknown_dims = [x for x in data_var_def.values() if x["dim_type"] not in ["2", "3", "4"]]
    if len(unknown_dims) > 0:
        raise ValueError(
            f"Invalid dimension specifications for {len(unknown_dims)} variables. Only supported are characters 2, 3, 4",
        )

    variables = {}
    variables["metadatavars"] = variables_metadata

    data_variables = empty_data_variables(data_var_def, time_dim, lead_time_dim, station_dim, ensemble_dim)
    variables["datavars"] = data_variables

    return variables


def empty_data_variables(
    data_var_def: dict,
    time_dim_tmp: Tuple[str, np.ndarray, Dict[str, str]],  # noqa: ARG001
    lead_time_dim_tmp: Tuple[str, np.ndarray, Dict[str, str]],  # noqa: ARG001
    station_dim_tmp: Tuple[str, np.ndarray, Dict[str, str]],  # noqa: ARG001
    ensemble_dim_tmp: Tuple[str, np.ndarray, Dict[str, str]],  # noqa: ARG001
) -> dict:
    """Create data variables as defined in the definition."""
    raise NotImplementedError("Not implemented yet")

    data_variables = {}

    ens_fcast_data_var_def = [x for x in data_var_def.values() if x["dim_type"] == "4"]
    ens_data_var_def = [x for x in data_var_def.values() if x["dim_type"] == "3"]
    point_data_var_def = [x for x in data_var_def.values() if x["dim_type"] == "2"]

    time_dim = "not implemented"
    lead_time_dim = "not implemented"
    station_dim = "not implemented"
    ensemble_dim = "not implemented"

    data_variables.update(
        {
            x["name"]: create_data_variable(
                x,
                [lead_time_dim, station_dim, ensemble_dim, time_dim],
            )
            for x in ens_fcast_data_var_def
        },
    )
    data_variables.update(
        {x["name"]: create_data_variable(x, [station_dim, ensemble_dim, time_dim]) for x in ens_data_var_def},
    )
    data_variables.update(
        {x["name"]: create_data_variable(x, [station_dim, time_dim]) for x in point_data_var_def},
    )

    return data_variables



import os
from typing import Optional

# import netCDF4
import numpy as np
import pandas as pd

from efts_io.dimensions import create_time_info, create_timestamps
from efts_io.wrapper import EftsDataSet
from efts_io.conventions import LOCATION_TYPE_ATTR_KEY, TYPE_ATTR_KEY, TYPE_DESCRIPTION_ATTR_KEY, UNITS_ATTR_KEY

pkg_dir = os.path.join(os.path.dirname(__file__), "..")

variable_names = ["variable_1", "variable_2"]
station_ids_ints = [123, 456]

nEns = 3
nLead = 4
x = np.arange(1, (nEns * nLead) + 1)
x = x.reshape((nLead, nEns))
y = x + nEns * nLead

timeAxisStart = pd.Timestamp(
    year=2010,
    month=8,
    day=1,
    hour=12,
    minute=0,
    second=0,
    tz="UTC",
)
tested_fcast_issue_time = timeAxisStart + pd.Timedelta(6, "h")
v1 = variable_names[0]
s1 = station_ids_ints[0]
v2 = variable_names[1]
s2 = station_ids_ints[1]


def dhours(i):
    return pd.Timedelta(i, "h")


def ddays(i):
    return pd.Timedelta(i * 24, "h")


import pytest


@pytest.mark.skip(reason="Ported from the R package, but may not be relevant or the best approach anymore")
def test_read_thing():
    fn = os.path.join(pkg_dir, "tests", "data", "hourly_test.nc")
    assert os.path.exists(fn)
    ds = EftsDataSet(fn)
    assert set(ds.get_dim_names()) == {"ens_member", "lead_time", "station", "str_len", "time"}
    r1 = ds.get_ensemble_forecasts(
        variable_name=v1,
        identifier=s1,
        start_time=tested_fcast_issue_time,
    )
    r2 = ds.get_ensemble_forecasts(
        variable_name=v2,
        identifier=s2,
        start_time=tested_fcast_issue_time,
    )
    assert r1[1, 1] == 6
    assert r2[1, 1] == 18
    # Check the lead time axix:
    # fcast_timeaxis = index(r1)
    # assert (fcast_timeaxis[0], tested_fcast_issue_time + lead_ts(lead_time_step_start_offset))
    # assert (fcast_timeaxis[1], tested_fcast_issue_time + lead_ts(lead_time_step_start_offset + lead_time_step_delta))


def _do_time_axis_test(
    tstart: pd.Timestamp,
    time_step: str = "days since",
    time_step_delta: int = 1,
    n: int = 10,
    tz_str: Optional[str] = None,
    expected_offset: Optional[pd.DateOffset] = None,
):
    time_dim_info = create_time_info(
        start=tstart,
        n=n,
        time_step=time_step,
        time_step_delta=time_step_delta,
    )
    timestamps = create_timestamps(time_dim_info, tz_str)
    expected_timestamps = [tstart + expected_offset * i for i in range(n)]
    assert np.all(timestamps == expected_timestamps)


def test_time_axis():
    n = 10
    tz_str = "UTC"
    tstart = pd.Timestamp(
        year=2010,
        month=8,
        day=1,
        hour=23,
        minute=0,
        second=0,
        tz=tz_str,
    )

    for time_step, time_step_delta, expected_offset in [
        ("hours since", 1, pd.DateOffset(hours=1)),
        ("hours since", 3, pd.DateOffset(hours=3)),
        ("days since", 1, pd.DateOffset(days=1)),
        ("days since", 3, pd.DateOffset(days=3)),
        # TODO
        # ("weeks since", 1, pd.DateOffset(weeks=1)),
        # ("months since", 1, pd.DateOffset(months=1)),
    ]:
        _do_time_axis_test(
            tstart,
            time_step,
            time_step_delta,
            n,
            tz_str=tz_str,
            expected_offset=expected_offset,
        )


# put tests in a tryCatch, to maximise the chances of cleaning up temporary
# files.
def doTests(
    tempNcFname,
    lead_time_tstep="hours",
    time_step="hours since",
    time_step_delta=1,
    lead_time_step_start_offset=1,
    lead_time_step_delta=1,
):
    # lead_time_tstep = "days"
    # time_step = "days since"
    # time_step_delta = 1L
    # lead_time_step_start_offset = 1L
    # lead_time_step_delta = 1L

    case_params = "".join(
        [
            "lts=",
            lead_time_tstep,
            ",ts=",
            time_step,
            ",tsdelta=",
            str(time_step_delta),
            ",ltsoffset=",
            str(lead_time_step_start_offset),
            ",ltsdelta=",
            str(lead_time_step_delta),
        ],
    )
    from efts_io.dimensions import create_time_info

    time_dim_info = create_time_info(
        start=timeAxisStart,
        n=10,
        time_step=time_step,
        time_step_delta=time_step_delta,
    )

    n = len(variable_names)
    varsDef = pd.DataFrame.from_dict(
        {
            "name": variable_names,
            "longname": ["long name for " + name for name in variable_names],
            UNITS_ATTR_KEY: np.repeat("mm", n),
            "missval": np.repeat(-999, n),
            "precision": np.repeat("double", n),
            TYPE_ATTR_KEY: np.repeat(2, n),
            "dimensions": np.repeat("4", n),
            TYPE_DESCRIPTION_ATTR_KEY: np.repeat("accumulated over the previous time step", n),
            LOCATION_TYPE_ATTR_KEY: np.repeat("Point", n),
        },
    )
    from efts_io.attributes import create_global_attributes

    glob_attr = create_global_attributes(
        title="title test",
        institution="test",
        source="test",
        catchment="dummy",
        comment="none",
    )

    from efts_io.variables import create_variable_definitions

    var_defs_dict = create_variable_definitions(varsDef)
    lead_times_offsets = (
        np.arange(lead_time_step_start_offset, lead_time_step_start_offset + nLead) * lead_time_step_delta
    )

    tz_str = "UTC"

    issue_times = create_timestamps(time_dim_info, tz_str)
    from efts_io.wrapper import xr_efts

    # TODO: expand to test non-integer station_ids
    station_ids = [str(i) for i in station_ids_ints]
    ensemble_size = nEns
    station_names = ["station_" + str(i) for i in station_ids_ints]
    xr_data = xr_efts(
        issue_times,
        station_ids,
        lead_times_offsets,
        lead_time_tstep,
        ensemble_size,
        station_names,
        nc_attributes=glob_attr,
    )

    # snc = create_efts(
    #     tempNcFname,
    #     time_dim_info,
    #     var_defs_dict,
    #     station_ids_ints,
    #     nc_attributes=glob_attr,
    #     lead_length=nLead,
    #     ensemble_length=nEns,
    #     lead_time_tstep=lead_time_tstep,
    # )
    snc = EftsDataSet(xr_data)

    snc.create_data_variables(var_defs_dict)

    snc.put_ensemble_forecasts(
        x,
        variable_name=v1,
        identifier=s1,
        start_time=tested_fcast_issue_time,
    )
    snc.put_ensemble_forecasts(
        y,
        variable_name=v2,
        identifier=s2,
        start_time=tested_fcast_issue_time,
    )

    r1 = snc.get_ensemble_forecasts(
        variable_name=v1,
        identifier=s1,
        start_time=tested_fcast_issue_time,
    )
    r2 = snc.get_ensemble_forecasts(
        variable_name=v2,
        identifier=s2,
        start_time=tested_fcast_issue_time,
    )
    assert r1[1, 1] == 6
    assert r2[1, 1] == 18
    snc.write()

    if lead_time_tstep == "hours":
        lead_ts = dhours
    elif lead_time_tstep == "days":
        lead_ts = ddays

    snc.to_netcdf(tempNcFname)

    from efts_io.wrapper import open_efts

    snc = open_efts(tempNcFname)
    r1 = snc.get_ensemble_forecasts(
        variable_name=v1,
        identifier=s1,
        start_time=tested_fcast_issue_time,
    )
    r2 = snc.get_ensemble_forecasts(
        variable_name=v2,
        identifier=s2,
        start_time=tested_fcast_issue_time,
    )
    assert r1[1, 1] == 6
    assert r2[1, 1] == 18
    # Check the lead time axix:
    fcast_timeaxis = r1.lead_time
    assert fcast_timeaxis[0] == tested_fcast_issue_time + lead_ts(
        lead_time_step_start_offset,
    )
    assert fcast_timeaxis[1] == tested_fcast_issue_time + lead_ts(
        lead_time_step_start_offset + lead_time_step_delta,
    )
    snc.close()


import tempfile
import pytest


@pytest.mark.skip(reason="Ported from the R package, but may not be relevant or the best approach anymore")
def test_round_trip():
    with tempfile.TemporaryDirectory() as temp_dir:
        tested_fcast_issue_time = timeAxisStart + ddays(2)
        # Covers https://github.com/jmp75/efts/issues/6
        tempNcFname = os.path.join(temp_dir, "days.nc")
        doTests(
            tempNcFname,
            lead_time_tstep="days",
            time_step="days since",
            time_step_delta=1,
            lead_time_step_start_offset=1,
            lead_time_step_delta=1,
        )

        tested_fcast_issue_time = timeAxisStart + dhours(6)

        tempNcFname = os.path.join(temp_dir, "hourly.nc")
        doTests(
            tempNcFname,
            lead_time_tstep="hours",
            time_step="hours since",
            time_step_delta=1,
            lead_time_step_start_offset=1,
            lead_time_step_delta=1,
        )

        tempNcFname = os.path.join(temp_dir, "three_hourly.nc")
        doTests(
            tempNcFname,
            lead_time_tstep="hours",
            time_step="hours since",
            time_step_delta=1,
            lead_time_step_start_offset=1,
            lead_time_step_delta=3,
        )


if __name__ == "__main__":
    # test_time_axis()
    # test_read_thing()
    test_round_trip()



"""Naming conventions for the EFTS netCDF file format."""

from datetime import datetime  # noqa: I001
from typing import Any, Dict, Iterable, List, Optional, Union

import numpy as np
import pandas as pd
import xarray as xr
from enum import Enum

# It may be important to import this AFTER xarray...
import netCDF4 as nc  # noqa: N813

ConvertibleToTimestamp = Union[str, datetime, np.datetime64, pd.Timestamp]
TYPES_CONVERTIBLE_TO_TIMESTAMP = [str, datetime, np.datetime64, pd.Timestamp]
"""Definition of a 'type' for type hints.
"""


TIME_DIMNAME = "time"
STATION_DIMNAME = "station"
ENS_MEMBER_DIMNAME = "ens_member"
LEAD_TIME_DIMNAME = "lead_time"
STR_LEN_DIMNAME = "strLen"

# New names for in-memory representation in an xarray way
# https://github.com/csiro-hydroinformatics/efts-io/issues/2
STATION_ID_DIMNAME = "station_id"
REALISATION_DIMNAME = "realisation"

# int station_id[station]
STATION_ID_VARNAME = "station_id"
# char station_name[str_len,station]
STATION_NAME_VARNAME = "station_name"
# float lat[station]
LAT_VARNAME = "lat"
# float lon[station]
LON_VARNAME = "lon"
# float x[station]
X_VARNAME = "x"
# float y[station]
Y_VARNAME = "y"
# float area[station]
AREA_VARNAME = "area"
# float elevation[station]
ELEVATION_VARNAME = "elevation"

conventional_varnames_mandatory = [
    STATION_DIMNAME,
    LEAD_TIME_DIMNAME,
    TIME_DIMNAME,
    ENS_MEMBER_DIMNAME,
    STR_LEN_DIMNAME,
    STATION_ID_VARNAME,
    STATION_NAME_VARNAME,
    LAT_VARNAME,
    LON_VARNAME,
]

conventional_varnames_optional = [
    X_VARNAME,
    Y_VARNAME,
    AREA_VARNAME,
    ELEVATION_VARNAME,
]

conventional_varnames = conventional_varnames_mandatory + conventional_varnames_optional

hydro_varnames = ("rain", "pet", "q", "swe", "tmin", "tmax")
var_type = ("obs", "sim")
obs_hydro_varnames = tuple(f"{var}_{var_type[0]}" for var in hydro_varnames)
sim_hydro_varnames = tuple(f"{var}_{var_type[1]}" for var in hydro_varnames)
obs_hydro_varnames_qul = tuple(f"{x}_qul" for x in obs_hydro_varnames)
sim_hydro_varnames_qul = tuple(f"{x}_qul" for x in sim_hydro_varnames)
known_hydro_varnames = obs_hydro_varnames + sim_hydro_varnames + obs_hydro_varnames_qul + sim_hydro_varnames_qul

# TODO: perhaps deal with the state variable names. But, is it used in practice?

TITLE_ATTR_KEY = "title"
INSTITUTION_ATTR_KEY = "institution"
SOURCE_ATTR_KEY = "source"
CATCHMENT_ATTR_KEY = "catchment"
STF_CONVENTION_VERSION_ATTR_KEY = "STF_convention_version"
STF_NC_SPEC_ATTR_KEY = "STF_nc_spec"
COMMENT_ATTR_KEY = "comment"
HISTORY_ATTR_KEY = "history"

TIME_STANDARD_ATTR_KEY = "time_standard"
STANDARD_NAME_ATTR_KEY = "standard_name"
LONG_NAME_ATTR_KEY = "long_name"
AXIS_ATTR_KEY = "axis"
UNITS_ATTR_KEY = "units"

FILLVALUE_ATTR_KEY = "_FillValue"
TYPE_ATTR_KEY = "type"
TYPE_DESCRIPTION_ATTR_KEY = "type_description"
DAT_TYPE_DESCRIPTION_ATTR_KEY = "dat_type_description"
DAT_TYPE_ATTR_KEY = "dat_type"
LOCATION_TYPE_ATTR_KEY = "location_type"

# We use a URL at a specific commit point, to be used as a file attribute.
# STF_2_0_URL = "https://github.com/csiro-hydroinformatics/efts/blob/d7d43a995fb5e459bcb894e09b7bb89de03e285c/docs/netcdf_for_water_forecasting.md"
# July 2025, set a new location/commit point:
STF_2_0_URL = "https://github.com/csiro-hydroinformatics/efts-io/blob/42ee35f0f019e9bad48b94914429476a7e8278dc/docs/netcdf_for_water_forecasting.md"


mandatory_global_attributes_xr = [
    TITLE_ATTR_KEY,
    INSTITUTION_ATTR_KEY,
    SOURCE_ATTR_KEY,
    CATCHMENT_ATTR_KEY,
    COMMENT_ATTR_KEY,
    HISTORY_ATTR_KEY,
]

mandatory_global_attributes = [
    TITLE_ATTR_KEY,
    INSTITUTION_ATTR_KEY,
    SOURCE_ATTR_KEY,
    CATCHMENT_ATTR_KEY,
    STF_CONVENTION_VERSION_ATTR_KEY,
    STF_NC_SPEC_ATTR_KEY,
    COMMENT_ATTR_KEY,
    HISTORY_ATTR_KEY,
]

mandatory_netcdf_dimensions = [TIME_DIMNAME, STATION_DIMNAME, LEAD_TIME_DIMNAME, STR_LEN_DIMNAME, ENS_MEMBER_DIMNAME]
mandatory_xarray_dimensions = [TIME_DIMNAME, STATION_ID_DIMNAME, LEAD_TIME_DIMNAME, REALISATION_DIMNAME]

# mappings to help automatic handling between stf and in memory dimensions
stf_to_xr_dims = {
    TIME_DIMNAME: TIME_DIMNAME,
    ENS_MEMBER_DIMNAME: REALISATION_DIMNAME,
    STATION_DIMNAME: STATION_ID_DIMNAME,
    LEAD_TIME_DIMNAME: LEAD_TIME_DIMNAME,
}

xr_to_stf_dims = {
    TIME_DIMNAME: TIME_DIMNAME,
    REALISATION_DIMNAME: ENS_MEMBER_DIMNAME,
    STATION_ID_DIMNAME: STATION_DIMNAME,
    LEAD_TIME_DIMNAME: LEAD_TIME_DIMNAME,
}

mandatory_varnames_xr = [
    TIME_DIMNAME,
    LEAD_TIME_DIMNAME,
    STATION_ID_VARNAME,
    STATION_NAME_VARNAME,
    REALISATION_DIMNAME,
    LAT_VARNAME,
    LON_VARNAME,
]


class AttributesErrorLevel(Enum):
    """Controls the behavior of variable attribute checking functions."""

    NONE = 1
    ERROR = 2
    # WARNING = 3


def get_default_dim_order() -> List[str]:
    """Default order of dimensions in the netCDF file.

    Returns:
        List[str]: dimension names: [lead_time, stations, ensemble_member, time]
    """
    return [
        LEAD_TIME_DIMNAME,
        STATION_DIMNAME,
        ENS_MEMBER_DIMNAME,
        TIME_DIMNAME,
    ]


def check_index_found(
    index_id: Optional[int],
    identifier: str,
    dimension_id: str,
) -> None:
    """Helper function to check that a value (index) was is indeed found in the dimension."""
    # return isinstance(index_id, np.int64)
    if index_id is None:
        raise ValueError(
            f"identifier '{identifier}' not found in the dimension '{dimension_id}'",
        )


# MdDatasetsType = Union[nc.Dataset, xr.Dataset, xr.DataArray]
MdDatasetsType = Union[xr.Dataset, xr.DataArray]


def _is_nc_dataset(d: Any) -> bool:
    return isinstance(d, nc.Dataset)


def _is_nc_variable(d: Any) -> bool:
    return isinstance(d, nc.Variable)


def _is_ncdf4_withattrs(d: Any) -> bool:
    return _is_nc_dataset(d) or _is_nc_variable(d)


def _has_required_dimensions(
    d: MdDatasetsType,
    mandatory_dimensions: Iterable[str],
) -> bool:
    if _is_nc_dataset(d):
        return set(d.dimensions.keys()) == set(mandatory_dimensions)
    import warnings

    with warnings.catch_warnings():
        warnings.simplefilter(action="ignore", category=FutureWarning)
        # FutureWarning: The return type of `Dataset.dims` will be changed
        # to return a set of dimension names in future, in order to be more
        # consistent with `DataArray.dims`.
        dims = d.dims
        # work around legacy discrepancy between data arrays and datasets: list and dict.
        kk = set([k for k in dims])  # noqa: C403, C416
        return kk == set(mandatory_dimensions)


def _is_subset_required_dimensions(
    d: MdDatasetsType,
    mandatory_dimensions: Iterable[str],
) -> bool:
    if _is_nc_dataset(d):
        d_set = set(d.dimensions.keys())
        return d_set.intersection(set(mandatory_dimensions)) == d_set
    import warnings

    with warnings.catch_warnings():
        warnings.simplefilter(action="ignore", category=FutureWarning)
        # FutureWarning: The return type of `Dataset.dims` will be changed
        # to return a set of dimension names in future, in order to be more
        # consistent with `DataArray.dims`.
        dims = d.dims
        # work around legacy discrepancy between data arrays and datasets: list and dict.
        d_set = set([k for k in dims])  # noqa: C403, C416
        return d_set.intersection(set(mandatory_dimensions)) == d_set


def has_required_stf2_dimensions(d: MdDatasetsType, mandatory_dimensions: Optional[Iterable[str]] = None) -> bool:
    """Has the dataset the required dimensions for STF conventions.

    Args:
        d (MdDatasetsType): data object to check

    Returns:
        bool: Has it the minimum STF dimentions
    """
    mandatory_dimensions = mandatory_dimensions or mandatory_netcdf_dimensions
    return _has_required_dimensions(d, mandatory_dimensions)


def has_required_xarray_dimensions(d: MdDatasetsType) -> bool:
    """Has the dataset the required dimensions for an in memory xarray representation."""
    return _has_required_dimensions(d, mandatory_xarray_dimensions)


def is_subset_required_xarray_dimensions(d: MdDatasetsType) -> bool:
    """Has the data array or dataset dimensions that are a subset of the spedified dims?"""
    return _is_subset_required_dimensions(d, mandatory_xarray_dimensions)


def _has_all_members(tested: Iterable[str], reference: Iterable[str]) -> bool:
    """Tests whether all the expected members are present in the tested set."""
    r = set(reference)
    return set(tested).intersection(r) == r


def has_required_global_attributes(d: MdDatasetsType) -> bool:
    """has_required_global_attributes."""
    if _is_nc_dataset(d):
        a = d.ncattrs()
        tested = set(a)
    else:
        a = d.attrs.keys()
        tested = set(a)
    return _has_all_members(tested, mandatory_global_attributes)


def has_required_xarray_global_attributes(d: MdDatasetsType) -> bool:
    """has_required_xarray_global_attributes."""
    a = d.attrs.keys()
    tested = set(a)
    return _has_all_members(tested, mandatory_global_attributes_xr)


def has_required_variables_xr(d: MdDatasetsType) -> bool:
    """has_required_variables."""
    a = d.variables.keys()
    tested = set(a)
    # Note: even if xarray, we do not need to check for the 'data_vars' attribute here.
    # a = d.data_vars.keys()
    # tested = set(a)
    return _has_all_members(tested, mandatory_varnames_xr)


def has_variable(d: MdDatasetsType, varname: str) -> bool:
    """has_variable."""
    a = d.variables.keys()
    tested = set(a)
    return varname in tested


def check_stf_compliance(file_path: str) -> Dict[str, List[str]]:
    """Checks the compliance of a netCDF file with the STF convention.

    Args:
        file_path (str): The path to the netCDF file.

    Returns:
        Dict[str, List[str]]: A dictionary with keys "INFO", "WARNING", "ERROR" and values as lists of strings describing compliance issues.
    """
    try:
        dataset = nc.Dataset(file_path, mode="r")
        results = {"INFO": [], "WARNING": [], "ERROR": []}

        # Check for required dimensions
        required_dims = [TIME_DIMNAME, STATION_DIMNAME, LEAD_TIME_DIMNAME, ENS_MEMBER_DIMNAME, STR_LEN_DIMNAME]
        available_dims = dataset.dimensions.keys()

        for dim in required_dims:
            if dim in available_dims:
                results["INFO"].append(f"Dimension '{dim}' is present.")
            else:
                results["ERROR"].append(f"Missing required dimension '{dim}'.")

        # Check global attributes
        required_global_attributes = [
            TITLE_ATTR_KEY,
            INSTITUTION_ATTR_KEY,
            SOURCE_ATTR_KEY,
            CATCHMENT_ATTR_KEY,
            STF_CONVENTION_VERSION_ATTR_KEY,
            STF_NC_SPEC_ATTR_KEY,
            COMMENT_ATTR_KEY,
            HISTORY_ATTR_KEY,
        ]
        available_global_attributes = dataset.ncattrs()

        for attr in required_global_attributes:
            if attr in available_global_attributes:
                results["INFO"].append(f"Global attribute '{attr}' is present.")
            else:
                results["WARNING"].append(f"Missing global attribute '{attr}'.")

        # Check mandatory variables and their attributes
        mandatory_variables = [
            TIME_DIMNAME,
            STATION_ID_VARNAME,
            STATION_NAME_VARNAME,
            ENS_MEMBER_DIMNAME,
            LEAD_TIME_DIMNAME,
            LAT_VARNAME,
            LON_VARNAME,
        ]
        variable_attributes = {
            TIME_DIMNAME: [
                STANDARD_NAME_ATTR_KEY,
                LONG_NAME_ATTR_KEY,
                UNITS_ATTR_KEY,
                TIME_STANDARD_ATTR_KEY,
                AXIS_ATTR_KEY,
            ],
            STATION_ID_VARNAME: [LONG_NAME_ATTR_KEY],
            STATION_NAME_VARNAME: [LONG_NAME_ATTR_KEY],
            ENS_MEMBER_DIMNAME: [STANDARD_NAME_ATTR_KEY, LONG_NAME_ATTR_KEY, UNITS_ATTR_KEY, AXIS_ATTR_KEY],
            LEAD_TIME_DIMNAME: [STANDARD_NAME_ATTR_KEY, LONG_NAME_ATTR_KEY, UNITS_ATTR_KEY, AXIS_ATTR_KEY],
            LAT_VARNAME: [LONG_NAME_ATTR_KEY, UNITS_ATTR_KEY, AXIS_ATTR_KEY],
            LON_VARNAME: [LONG_NAME_ATTR_KEY, UNITS_ATTR_KEY, AXIS_ATTR_KEY],
        }

        for var in mandatory_variables:
            if var in dataset.variables:
                results["INFO"].append(f"Mandatory variable '{var}' is present.")
                # Check attributes
                for attr, required_attrs in variable_attributes.items():
                    if var == attr:
                        for req_attr in required_attrs:
                            if req_attr in dataset.variables[var].ncattrs():
                                results["INFO"].append(f"Attribute '{req_attr}' for variable '{var}' is present.")
                            else:
                                results["WARNING"].append(
                                    f"Missing required attribute '{req_attr}' for variable '{var}'.",
                                )
            else:
                results["ERROR"].append(f"Missing mandatory variable '{var}'.")

        dataset.close()
        return results  # noqa: TRY300

    except Exception as e:  # noqa: BLE001
        return {"ERROR": [f"Error opening file '{file_path}': {e!s}"]}


def _is_structural_varname(name: str) -> bool:
    return name in conventional_varnames


def _is_known_hydro_varname(name: str) -> bool:
    """Checks if the variable name is a known hydrologic variable."""
    # TODO: perhaps deal with state variable conventional names.
    return name in known_hydro_varnames


def _is_observation_variable(name: str) -> bool:
    return name in obs_hydro_varnames


def _is_simulation_variable(name: str) -> bool:
    return name in sim_hydro_varnames


def _is_quality_variable(name: str) -> bool:
    return name in obs_hydro_varnames_qul or name in sim_hydro_varnames_qul


def _extract_var_type(variable: Any) -> str:
    if _is_observation_variable(variable):
        return "obs"
    if _is_simulation_variable(variable):
        return "sim"
    if _is_quality_variable(variable):
        return "qul"
    return None


def _check_variable_attributes_obs(
    variable: Any,
    error_threshold: AttributesErrorLevel = AttributesErrorLevel.NONE,
) -> List[str]:
    """Checks if the attributes of the observed variable comply with the conventions."""
    missing_attributes_messages = []
    required_attributes = {
        LONG_NAME_ATTR_KEY: str,
        UNITS_ATTR_KEY: str,
        FILLVALUE_ATTR_KEY: float,
        TYPE_ATTR_KEY: int,
        TYPE_DESCRIPTION_ATTR_KEY: str,
        DAT_TYPE_ATTR_KEY: str,
        LOCATION_TYPE_ATTR_KEY: str,
    }
    return _check_attrs(variable, required_attributes, missing_attributes_messages, error_threshold=error_threshold)


def _template_variable_attributes():  # noqa: ANN202
    return {
        LONG_NAME_ATTR_KEY: "",
        UNITS_ATTR_KEY: "",
        FILLVALUE_ATTR_KEY: -9999.0,
        TYPE_ATTR_KEY: 0,
        TYPE_DESCRIPTION_ATTR_KEY: "",
        DAT_TYPE_ATTR_KEY: "",
        LOCATION_TYPE_ATTR_KEY: "Point",
    }


def _check_variable_attributes_sim(
    variable: Any,
    error_threshold: AttributesErrorLevel = AttributesErrorLevel.NONE,
) -> List[str]:
    """Checks if the attributes of the simulated variable comply with the conventions."""
    missing_attributes_messages = []
    required_attributes = {
        LONG_NAME_ATTR_KEY: str,
        UNITS_ATTR_KEY: str,
        FILLVALUE_ATTR_KEY: float,
        TYPE_ATTR_KEY: int,
        TYPE_DESCRIPTION_ATTR_KEY: str,
        DAT_TYPE_ATTR_KEY: str,
        LOCATION_TYPE_ATTR_KEY: str,
    }
    return _check_attrs(variable, required_attributes, missing_attributes_messages, error_threshold=error_threshold)


def _check_variable_attributes_qul(
    variable: Any,
    error_threshold: AttributesErrorLevel = AttributesErrorLevel.NONE,
) -> List[str]:
    """Checks if the attributes of the data quality code variable comply with the conventions."""
    missing_attributes_messages = []
    required_attributes = {
        LONG_NAME_ATTR_KEY: str,
        UNITS_ATTR_KEY: str,
        FILLVALUE_ATTR_KEY: int,
        LOCATION_TYPE_ATTR_KEY: str,
        TYPE_DESCRIPTION_ATTR_KEY: str,
        DAT_TYPE_ATTR_KEY: str,
    }
    return _check_attrs(variable, required_attributes, missing_attributes_messages, error_threshold=error_threshold)


def _check_attrs_ncdataset(
    variable: Any,
    required_attributes: Dict[str, type],
    missing_attributes_messages: List[str],
    error_threshold: AttributesErrorLevel = AttributesErrorLevel.NONE,
) -> List[str]:
    for attr, attr_type in required_attributes.items():
        if attr not in variable.ncattrs():
            missing_attributes_messages.append(f"Missing required attribute '{attr}' for variable '{variable.name}'.")
        else:
            actual_type = type(variable.getncattr(attr))
            if actual_type != attr_type:
                missing_attributes_messages.append(
                    f"Attribute '{attr}' for variable '{variable.name}' has an unexpected type '{actual_type.__name__}'. Expected type: '{attr_type.__name__}'.",
                )
    if error_threshold == AttributesErrorLevel.ERROR and missing_attributes_messages:
        raise ValueError(
            f"Variable '{variable.name}' has missing or incorrect attributes: {missing_attributes_messages}",
        )
    return missing_attributes_messages


def _check_attrs_xr(
    variable: MdDatasetsType,
    required_attributes: Dict[str, type],
    missing_attributes_messages: List[str],
    error_threshold: AttributesErrorLevel = AttributesErrorLevel.NONE,
) -> List[str]:
    for attr, attr_type in required_attributes.items():
        if attr not in variable.attrs:
            missing_attributes_messages.append(f"Missing required attribute '{attr}' for variable '{variable.name}'.")
        else:
            actual_type = type(variable.attrs[attr])
            if actual_type != attr_type:
                missing_attributes_messages.append(
                    f"Attribute '{attr}' for variable '{variable.name}' has an unexpected type '{actual_type.__name__}'. Expected type: '{attr_type.__name__}'.",
                )
    if error_threshold == AttributesErrorLevel.ERROR and missing_attributes_messages:
        raise ValueError(
            f"Variable '{variable.name}' has missing or incorrect attributes: {missing_attributes_messages}",
        )
    return missing_attributes_messages


def _check_attrs(
    variable: Any,
    required_attributes: Dict[str, type],
    missing_attributes_messages: List[str],
    error_threshold: AttributesErrorLevel = AttributesErrorLevel.NONE,
) -> List[str]:
    if _is_ncdf4_withattrs(variable):
        return _check_attrs_ncdataset(variable, required_attributes, missing_attributes_messages, error_threshold)
    else:  # noqa: RET505
        return _check_attrs_xr(variable, required_attributes, missing_attributes_messages, error_threshold)


def _check_variable_attributes(variable: Any) -> List[str]:
    """Checks if the attributes of a variable comply with the conventions depending on the type of variable.

    Args:
        variable (Any): The netCDF variable whose attributes are to be checked.

    Returns:
        List[str]: A list of messages describing any missing attributes.
    """
    var_type = _extract_var_type(variable.name)

    if var_type == "obs":
        return _check_variable_attributes_obs(variable)
    if var_type == "sim":
        return _check_variable_attributes_sim(variable)
    if var_type == "qul":
        return _check_variable_attributes_qul(variable)

    return []


def check_hydrologic_variables(file_path: str) -> Dict[str, List[str]]:
    """Checks if the variable names and attributes in a netCDF file comply with the STF convention.

    Args:
        file_path (str): The path to the netCDF file.

    Returns:
        Dict[str, List[str]]: A dictionary with keys "INFO", "WARNING", "ERROR" and values as lists of strings describing compliance issues.
    """
    try:
        dataset = None
        dataset = nc.Dataset(file_path, mode="r")
        results = {"INFO": [], "WARNING": [], "ERROR": []}

        for var in dataset.variables:
            if _is_structural_varname(var):
                continue
            if _is_known_hydro_varname(var):
                results["INFO"].append(f"Hydrologic variable '{var}' follows the recommended naming convention.")

                # Check attributes
                for msg in _check_variable_attributes(dataset.variables[var]):
                    results["WARNING"].append(msg)
            else:
                results["WARNING"].append(
                    f"Hydrologic variable '{var}' does not follow the recommended naming convention.",
                )

        return results  # noqa: TRY300

    except Exception as e:  # noqa: BLE001
        return {"ERROR": [f"Error opening or reading file '{file_path}': {e!s}"]}

    finally:
        if dataset:
            dataset.close()


def check_optional_variable_attributes(
    variable: Any,
    error_threshold: AttributesErrorLevel = AttributesErrorLevel.NONE,
) -> List[str]:
    """Checks if the attributes of the observed variable comply with the conventions."""
    missing_attributes_messages = []
    required_attributes = {
        STANDARD_NAME_ATTR_KEY: str,
        LONG_NAME_ATTR_KEY: str,
        UNITS_ATTR_KEY: str,
    }
    return _check_attrs(variable, required_attributes, missing_attributes_messages, error_threshold=error_threshold)


def convert_to_datetime64_utc(x: ConvertibleToTimestamp) -> np.datetime64:
    """Converts a known timestamp representation an np.datetime64."""
    if isinstance(x, pd.Timestamp):
        if x.tz is None:
            x = x.tz_localize("UTC")
        x = x.tz_convert("UTC")
    elif isinstance(x, datetime):
        x = pd.Timestamp(x, tz="UTC") if x.tzinfo is None else pd.Timestamp(x).tz_convert("UTC")
    elif isinstance(x, str):
        x_dt = pd.to_datetime(x)
        x = pd.Timestamp(x_dt, tz="UTC") if x_dt.tzinfo is None else pd.Timestamp(x_dt).tz_convert("UTC")
    elif isinstance(x, np.datetime64):
        x = pd.Timestamp(x).tz_localize("UTC")
    else:
        raise TypeError(f"Cannot convert {type(x)} to np.datetime64 with UTC timezone.")

    return x.to_datetime64()


def exportable_to_stf2(data: MdDatasetsType) -> bool:
    """Check if the dataset can be written to a netCDF file compliant with STF 2.0 specification.

    This method checks if the underlying xarray dataset or dataarray has the required dimensions and global attributes as specified by the STF 2.0 convention.

    Returns:
        bool: True if the dataset can be written to a STF 2.0 compliant netCDF file, False otherwise.
    """
    from efts_io.conventions import has_required_stf2_dimensions, has_required_variables_xr, mandatory_xarray_dimensions  # noqa: I001

    required_stf2_dimensions = has_required_stf2_dimensions(data, mandatory_xarray_dimensions)
    required_attributes = has_required_xarray_global_attributes(data)
    required_variables = has_required_variables_xr(data)
    # Check that station_ids are not strings though:
    if STATION_ID_DIMNAME not in data:  # must be because of above checks, but no harm in checking
        return False
    station_ids = data[STATION_ID_DIMNAME].values
    # it can be an object type of string or integer, so let's check:
    supported_types = (np.integer, np.bytes_, np.str_)
    if not issubclass(station_ids.dtype.type, supported_types):
        return False

    return required_stf2_dimensions and required_attributes and required_variables
